from flask import abort
from flask_restful import Resource
from flask_jwt_extended import jwt_required


from zou.app.services import budget_service
from zou.app.mixin import ArgsMixin
from zou.app.services import (
    persons_service,
    projects_service,
    schedule_service,
    tasks_service,
    time_spents_service,
    user_service,
)
from zou.app.utils import permissions
from zou.app.services.exception import (
    WrongParameterException,
    WrongDateFormatException,
)
from zou.app.models.metadata_descriptor import METADATA_DESCRIPTOR_TYPES


class OpenProjectsResource(Resource, ArgsMixin):
    """
    Return the list of projects currently running. Most of the time, past
    projects are not needed.
    """

    @jwt_required()
    def get(self):
        """
        Get open projects
        ---
        description: Return the list of projects currently running. Most of the
          time, past projects are not needed.
        tags:
          - Projects
        parameters:
          - in: query
            name: name
            required: false
            schema:
              type: string
            description: Filter projects by name
            example: "My Project"
        responses:
          200:
            description: All running projects
            content:
              application/json:
                schema:
                  type: array
                  items:
                    type: object
                    properties:
                      id:
                        type: string
                        format: uuid
                        description: Project unique identifier
                        example: a24a6ea4-ce75-4665-a070-57453082c25
                      name:
                        type: string
                        description: Project name
                        example: "My Project"
                      project_status_id:
                        type: string
                        format: uuid
                        description: Project status unique identifier
                        example: b35b7fb5-df86-5776-b181-68564193d36
        """
        name = self.get_text_parameter("name")
        if permissions.has_admin_permissions():
            return projects_service.open_projects(name)
        else:
            return user_service.get_open_projects(name)


class AllProjectsResource(Resource, ArgsMixin):
    """
    Return all projects listed in database. Ensure that user has at least
    the manager level before that.
    """

    @jwt_required()
    def get(self):
        """
        Get all projects
        ---
        description: Return all projects listed in database. Ensure that user has
          at least the manager level before that.
        tags:
          - Projects
        parameters:
          - in: query
            name: name
            required: false
            schema:
              type: string
            description: Filter projects by name
            example: "My Project"
        responses:
          200:
            description: All projects listed in database
            content:
              application/json:
                schema:
                  type: array
                  items:
                    type: object
                    properties:
                      id:
                        type: string
                        format: uuid
                        description: Project unique identifier
                        example: a24a6ea4-ce75-4665-a070-57453082c25
                      name:
                        type: string
                        description: Project name
                        example: "My Project"
                      project_status_id:
                        type: string
                        format: uuid
                        description: Project status unique identifier
                        example: b35b7fb5-df86-5776-b181-68564193d36
        """
        name = self.get_text_parameter("name")
        try:
            permissions.check_admin_permissions()

            if name is None:
                return projects_service.get_projects()
            else:
                return [projects_service.get_project_by_name(name)]
        except permissions.PermissionDenied:
            if name is None:
                return user_service.get_projects()
            else:
                return [user_service.get_project_by_name(name)]


class ProductionTeamResource(Resource, ArgsMixin):
    """
    Allow to manage the people listed in a production team.
    """

    @jwt_required()
    def get(self, project_id):
        """
        Get production team
        ---
        description: Return the people listed in a production team.
        tags:
          - Projects
        parameters:
          - in: path
            name: project_id
            required: true
            schema:
              type: string
              format: uuid
            description: Project unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
          200:
            description: People listed in a production team
            content:
              application/json:
                schema:
                  type: array
                  items:
                    type: object
                    properties:
                      id:
                        type: string
                        format: uuid
                        description: Person unique identifier
                        example: a24a6ea4-ce75-4665-a070-57453082c25
                      first_name:
                        type: string
                        description: Person first name
                        example: "John"
                      last_name:
                        type: string
                        description: Person last name
                        example: "Doe"
                      email:
                        type: string
                        description: Person email address
                        example: "john.doe@example.com"
        """
        user_service.check_project_access(project_id)
        project = projects_service.get_project_raw(project_id)
        persons = []
        for person in project.team:
            if permissions.has_manager_permissions():
                persons.append(person.serialize_safe(relations=True))
            else:
                persons.append(person.present_minimal())
        return persons

    @jwt_required()
    def post(self, project_id):
        """
        Add person to production team
        ---
        description: Add a person to a production team.
        tags:
          - Projects
        parameters:
          - in: path
            name: project_id
            required: true
            schema:
              type: string
              format: uuid
            description: Project unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: object
                required:
                  - person_id
                properties:
                  person_id:
                    type: string
                    format: uuid
                    description: Person unique identifier
                    example: b35b7fb5-df86-5776-b181-68564193d36
        responses:
          201:
            description: Person added to the production team
            content:
              application/json:
                schema:
                  type: object
                  properties:
                    id:
                      type: string
                      format: uuid
                      description: Project unique identifier
                      example: a24a6ea4-ce75-4665-a070-57453082c25
                    name:
                      type: string
                      description: Project name
                      example: "My Project"
          400:
            description: Invalid parameters
        """
        args = self.get_args([("person_id", "", True)])

        user_service.check_manager_project_access(project_id)
        return (
            projects_service.add_team_member(project_id, args["person_id"]),
            201,
        )


class ProductionTeamRemoveResource(Resource):

    @jwt_required()
    def delete(self, project_id, person_id):
        """
        Remove person from production team
        ---
        description: Remove people listed in a production team.
        tags:
          - Projects
        parameters:
          - in: path
            name: project_id
            required: true
            schema:
              type: string
              format: uuid
            description: Project unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: person_id
            required: true
            schema:
              type: string
              format: uuid
            description: Person unique identifier
            example: b35b7fb5-df86-5776-b181-68564193d36
        responses:
          204:
            description: Person removed from production team
        """
        user_service.check_manager_project_access(project_id)
        projects_service.remove_team_member(project_id, person_id)
        return "", 204


class ProductionAssetTypeResource(Resource, ArgsMixin):
    """
    Allow to add an asset type linked to a production.
    """

    @jwt_required()
    def post(self, project_id):
        """
        Add asset type to production
        ---
        description: Add an asset type linked to a production.
        tags:
          - Projects
        parameters:
          - in: path
            name: project_id
            required: true
            schema:
              type: string
              format: uuid
            description: Project unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: object
                required:
                  - asset_type_id
                properties:
                  asset_type_id:
                    type: string
                    format: uuid
                    description: Asset type unique identifier
                    example: b35b7fb5-df86-5776-b181-68564193d36
        responses:
          201:
            description: Asset type added to production
            content:
              application/json:
                schema:
                  type: object
                  properties:
                    id:
                      type: string
                      format: uuid
                      description: Project unique identifier
                      example: a24a6ea4-ce75-4665-a070-57453082c25
                    name:
                      type: string
                      description: Project name
                      example: "My Project"
          400:
            description: Invalid parameters
        """
        args = self.get_args([("asset_type_id", "", True)])

        user_service.check_manager_project_access(project_id)
        project = projects_service.add_asset_type_setting(
            project_id, args["asset_type_id"]
        )
        return project, 201


class ProductionAssetTypeRemoveResource(Resource):

    @jwt_required()
    def delete(self, project_id, asset_type_id):
        """
        Remove asset type from production
        ---
        description: Remove an asset type from a production.
        tags:
          - Projects
        parameters:
          - in: path
            name: project_id
            required: true
            schema:
              type: string
              format: uuid
            description: Project unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: asset_type_id
            required: true
            schema:
              type: string
              format: uuid
            description: Asset type unique identifier
            example: b35b7fb5-df86-5776-b181-68564193d36
        responses:
          204:
            description: Asset type removed from production
        """
        user_service.check_manager_project_access(project_id)
        projects_service.remove_asset_type_setting(project_id, asset_type_id)
        return "", 204


class ProductionTaskTypesResource(Resource, ArgsMixin):
    """
    Retrieve task types linked to the production
    """

    @jwt_required()
    def get(self, project_id):
        """
        Get production task types
        ---
        description: Retrieve task types linked to the production.
        tags:
          - Projects
        parameters:
          - in: path
            name: project_id
            required: true
            schema:
              type: string
              format: uuid
            description: Project unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
          200:
            description: Task types linked to the production
            content:
              application/json:
                schema:
                  type: array
                  items:
                    type: object
        """
        user_service.check_project_access(project_id)
        return projects_service.get_project_task_types(project_id)


class ProductionTaskTypeResource(Resource, ArgsMixin):
    """
    Allow to add a task type linked to a production.
    """

    @jwt_required()
    def post(self, project_id):
        """
        Add task type to production
        ---
        description: Add a task type linked to a production.
        tags:
          - Projects
        parameters:
          - in: path
            name: project_id
            required: true
            schema:
              type: string
              format: uuid
            description: Project unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: object
                required:
                  - task_type_id
                properties:
                  task_type_id:
                    type: string
                    format: uuid
                    description: Task type unique identifier
                    example: b35b7fb5-df86-5776-b181-68564193d36
                  priority:
                    type: string
                    description: Task type priority
                    example: "None"
        responses:
          201:
            description: Task type added to production
            content:
              application/json:
                schema:
                  type: object
                  properties:
                    id:
                      type: string
                      format: uuid
                      description: Project unique identifier
                      example: a24a6ea4-ce75-4665-a070-57453082c25
                    name:
                      type: string
                      description: Project name
                      example: "My Project"
          400:
            description: Invalid parameters
        """
        args = self.get_args(
            [("task_type_id", "", True), ("priority", None, False)]
        )

        user_service.check_manager_project_access(project_id)
        project = projects_service.add_task_type_setting(
            project_id, args["task_type_id"], args["priority"]
        )
        return project, 201


class ProductionTaskTypeRemoveResource(Resource):
    """
    Allow to remove a task type linked to a production.
    """

    @jwt_required()
    def delete(self, project_id, task_type_id):
        """
        Remove task type from production
        ---
        description: Remove a task type from a production.
        tags:
          - Projects
        parameters:
          - in: path
            name: project_id
            required: true
            schema:
              type: string
              format: uuid
            description: Project unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: task_type_id
            required: true
            schema:
              type: string
              format: uuid
            description: Task type unique identifier
            example: b35b7fb5-df86-5776-b181-68564193d36
        responses:
          204:
            description: Task type removed from production
        """
        user_service.check_manager_project_access(project_id)
        projects_service.remove_task_type_setting(project_id, task_type_id)
        return "", 204


class ProductionTaskStatusResource(Resource, ArgsMixin):
    """
    Allow to add a task type linked to a production.
    """

    @jwt_required()
    def get(self, project_id):
        """
        Get production task statuses
        ---
        description: Return task statuses linked to a production.
        tags:
          - Projects
        parameters:
          - in: path
            name: project_id
            required: true
            schema:
              type: string
              format: uuid
            description: Project unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
          200:
            description: Task statuses linked to production
            content:
              application/json:
                schema:
                  type: array
                  items:
                    type: object
        """
        user_service.check_project_access(project_id)
        return projects_service.get_project_task_statuses(project_id)

    @jwt_required()
    def post(self, project_id):
        """
        Add task status to production
        ---
        description: Add a task status linked to a production.
        tags:
          - Projects
        parameters:
          - in: path
            name: project_id
            required: true
            schema:
              type: string
              format: uuid
            description: Project unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: object
                required:
                  - task_status_id
                properties:
                  task_status_id:
                    type: string
                    format: uuid
                    description: Task status unique identifier
                    example: b35b7fb5-df86-5776-b181-68564193d36
        responses:
          201:
            description: Task status added to production
            content:
              application/json:
                schema:
                  type: object
                  properties:
                    id:
                      type: string
                      format: uuid
                      description: Project unique identifier
                      example: a24a6ea4-ce75-4665-a070-57453082c25
                    name:
                      type: string
                      description: Project name
                      example: "My Project"
          400:
            description: Invalid parameters
        """
        args = self.get_args([("task_status_id", "", True)])

        user_service.check_manager_project_access(project_id)
        project = projects_service.add_task_status_setting(
            project_id, args["task_status_id"]
        )
        return project, 201


class ProductionTaskStatusRemoveResource(Resource):
    """
    Allow to remove a task status linked to a production.
    """

    @jwt_required()
    def delete(self, project_id, task_status_id):
        """
        Remove task status from production
        ---
        description: Remove a task status from a production.
        tags:
          - Projects
        parameters:
          - in: path
            name: project_id
            required: true
            schema:
              type: string
              format: uuid
            description: Project unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: task_status_id
            required: true
            schema:
              type: string
              format: uuid
            description: Task status unique identifier
            example: b35b7fb5-df86-5776-b181-68564193d36
        responses:
          204:
            description: Task status removed from production
        """
        user_service.check_manager_project_access(project_id)
        projects_service.remove_task_status_setting(project_id, task_status_id)
        return "", 204


class ProductionStatusAutomationResource(Resource, ArgsMixin):
    """
    Allow to add a status automation linked to a production.
    """

    @jwt_required()
    def get(self, project_id):
        """
        Get production status automations
        ---
        description: Get status automations linked to a production.
        tags:
          - Projects
        parameters:
          - in: path
            name: project_id
            required: true
            schema:
              type: string
              format: uuid
            description: Project unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
          200:
            description: Status automations linked to production
            content:
              application/json:
                schema:
                  type: array
                  items:
                    type: object
        """
        user_service.check_manager_project_access(project_id)
        return projects_service.get_project_status_automations(project_id)

    @jwt_required()
    def post(self, project_id):
        """
        Add status automation to production
        ---
        description: Add a status automation linked to a production.
        tags:
          - Projects
        parameters:
          - in: path
            name: project_id
            required: true
            schema:
              type: string
              format: uuid
            description: Project unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: object
                required:
                  - status_automation_id
                properties:
                  status_automation_id:
                    type: string
                    format: uuid
                    description: Status automation unique identifier
                    example: b35b7fb5-df86-5776-b181-68564193d36
        responses:
          201:
            description: Status automation added to production
            content:
              application/json:
                schema:
                  type: object
                  properties:
                    id:
                      type: string
                      format: uuid
                      description: Project unique identifier
                      example: a24a6ea4-ce75-4665-a070-57453082c25
                    name:
                      type: string
                      description: Project name
                      example: "My Project"
        """
        args = self.get_args([("status_automation_id", "", True)])

        user_service.check_manager_project_access(project_id)
        project = projects_service.add_status_automation_setting(
            project_id, args["status_automation_id"]
        )
        return project, 201


class ProductionStatusAutomationRemoveResource(Resource):
    """
    Allow to remove a status automation linked to a production.
    """

    @jwt_required()
    def delete(self, project_id, status_automation_id):
        """
        Remove status automation from production
        ---
        description: Remove a status automation from a production.
        tags:
          - Projects
        parameters:
          - in: path
            name: project_id
            required: true
            schema:
              type: string
              format: uuid
            description: Project unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: status_automation_id
            required: true
            schema:
              type: string
              format: uuid
            description: Status automation unique identifier
            example: b35b7fb5-df86-5776-b181-68564193d36
        responses:
          204:
            description: Status automation removed from production
        """
        user_service.check_manager_project_access(project_id)
        projects_service.remove_status_automation_setting(
            project_id, status_automation_id
        )
        return "", 204


class ProductionPreviewBackgroundFileResource(Resource, ArgsMixin):
    """
    Allow to add a preview background file linked to a production.
    """

    @jwt_required()
    def get(self, project_id):
        """
        Get production preview background files
        ---
        description: Return preview background files linked to a production.
        tags:
          - Projects
        parameters:
          - in: path
            name: project_id
            required: true
            schema:
              type: string
              format: uuid
            description: Project unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
          200:
            description: Preview background files linked to production
            content:
              application/json:
                schema:
                  type: array
                  items:
                    type: object
        """
        user_service.check_project_access(project_id)
        return projects_service.get_project_preview_background_files(
            project_id
        )

    @jwt_required()
    def post(self, project_id):
        """
        Add preview background file to production
        ---
        description: Add a preview background file linked to a production.
        tags:
          - Projects
        parameters:
          - in: path
            name: project_id
            required: true
            schema:
              type: string
              format: uuid
            description: Project unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: object
                required:
                  - preview_background_file_id
                properties:
                  preview_background_file_id:
                    type: string
                    format: uuid
                    description: Preview background file unique identifier
                    example: b35b7fb5-df86-5776-b181-68564193d36
        responses:
          201:
            description: Preview background file added to production
            content:
              application/json:
                schema:
                  type: object
                  properties:
                    id:
                      type: string
                      format: uuid
                      description: Project unique identifier
                      example: a24a6ea4-ce75-4665-a070-57453082c25
                    name:
                      type: string
                      description: Project name
                      example: "My Project"
        """
        args = self.get_args([("preview_background_file_id", "", True)])

        user_service.check_manager_project_access(project_id)
        project = projects_service.add_preview_background_file_setting(
            project_id, args["preview_background_file_id"]
        )
        return project, 201


class ProductionPreviewBackgroundFileRemoveResource(Resource):
    """
    Allow to remove a preview background file linked to a production.
    """

    @jwt_required()
    def delete(self, project_id, preview_background_file_id):
        """
        Remove preview background file from production
        ---
        description: Remove a preview background file from a production.
        tags:
          - Projects
        parameters:
          - in: path
            name: project_id
            required: true
            schema:
              type: string
              format: uuid
            description: Project unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: preview_background_file_id
            required: true
            schema:
              type: string
              format: uuid
            description: Preview background file unique identifier
            example: b35b7fb5-df86-5776-b181-68564193d36
        responses:
          204:
            description: Preview background file removed from production
        """
        user_service.check_manager_project_access(project_id)
        projects_service.remove_preview_background_file_setting(
            project_id, preview_background_file_id
        )
        return "", 204


class ProductionMetadataDescriptorsResource(Resource, ArgsMixin):
    """
    Resource to get and create metadata descriptors. It serves to describe
    extra fields listed in the data attribute of entities.
    """

    @jwt_required()
    def get(self, project_id):
        """
        Get metadata descriptors
        ---
        description: Get all metadata descriptors. It serves to describe extra
          fields listed in the data attribute of entities.
        tags:
          - Projects
        parameters:
          - in: path
            name: project_id
            required: true
            schema:
              type: string
              format: uuid
            description: Project unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
          200:
            description: All metadata descriptors
            content:
              application/json:
                schema:
                  type: array
                  items:
                    type: object
        """
        user_service.check_project_access(project_id)
        for_client = permissions.has_client_permissions()
        return projects_service.get_metadata_descriptors(
            project_id, for_client
        )

    @jwt_required()
    def post(self, project_id):
        """
        Create metadata descriptor
        ---
        description: Create a new metadata descriptor. It serves to describe
          extra fields listed in the data attribute of entities.
        tags:
          - Projects
        parameters:
          - in: path
            name: project_id
            required: true
            schema:
              type: string
              format: uuid
            description: Project unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: object
                required:
                  - name
                  - data_type
                properties:
                  entity_type:
                    type: string
                    description: Entity type for the metadata descriptor
                    enum: ["Asset", "Shot", "Edit", "Episode", "Sequence"]
                    default: "Asset"
                    example: "Asset"
                  name:
                    type: string
                    description: Name of the metadata descriptor
                    example: "Custom Field"
                  data_type:
                    type: string
                    description: Type of data (string, number, boolean, etc.)
                    example: "string"
                  for_client:
                    type: string
                    description: Whether the descriptor is for client
                    default: "False"
                    example: "True"
                  choices:
                    type: array
                    description: List of choices for the descriptor
                    items:
                      type: string
                    example: ["option1", "option2"]
                  departments:
                    type: array
                    description: List of departments for the descriptor
                    items:
                      type: string
                    example: ["department1", "department2"]
        responses:
          201:
            description: Metadata descriptor created
            content:
              application/json:
                schema:
                  type: object
                  properties:
                    id:
                      type: string
                      format: uuid
                      description: Metadata descriptor unique identifier
                      example: a24a6ea4-ce75-4665-a070-57453082c25
                    name:
                      type: string
                      description: Metadata descriptor name
                      example: "Custom Field"
                    data_type:
                      type: string
                      description: Metadata descriptor data type
                      example: "string"
          400:
            description: Invalid parameters
        """
        args = self.get_args(
            [
                ("entity_type", "Asset", False),
                ("name", "", True),
                ("data_type", "string", True),
                ("for_client", "False", False),
                ("choices", [], False, str, "append"),
                ("departments", [], False, str, "append"),
            ]
        )

        user_service.check_all_departments_access(
            project_id, args["departments"]
        )

        args["for_client"] = args["for_client"] == "True"

        if args["entity_type"] not in [
            "Asset",
            "Shot",
            "Edit",
            "Episode",
            "Sequence",
        ]:
            raise WrongParameterException(
                "Wrong entity type. Please select Asset, Shot, Sequence, Episode or Edit."
            )

        if len(args["name"]) == 0:
            raise WrongParameterException("Name cannot be empty.")

        types = [type_name for type_name, _ in METADATA_DESCRIPTOR_TYPES]
        if args["data_type"] not in types:
            raise WrongParameterException("Invalid data_type")

        return (
            projects_service.add_metadata_descriptor(
                project_id,
                args["entity_type"],
                args["name"],
                args["data_type"],
                args["choices"],
                args["for_client"],
                args["departments"],
            ),
            201,
        )


class ProductionMetadataDescriptorResource(Resource, ArgsMixin):
    """
    Resource to get, update or delete a metadata descriptor. Descriptors serve
    to describe extra fields listed in the data attribute of entities.
    """

    @jwt_required()
    def get(self, project_id, descriptor_id):
        """
        Get metadata descriptor
        ---
        description: Get a metadata descriptor. Descriptors serve to describe
          extra fields listed in the data attribute of entities.
        tags:
          - Projects
        parameters:
          - in: path
            name: project_id
            required: true
            schema:
              type: string
              format: uuid
            description: Project unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: descriptor_id
            required: true
            schema:
              type: string
              format: uuid
            description: Metadata descriptor unique identifier
            example: b35b7fb5-df86-5776-b181-68564193d36
        responses:
          200:
            description: Metadata descriptor
            content:
              application/json:
                schema:
                  type: object
        """
        user_service.check_project_access(project_id)
        return projects_service.get_metadata_descriptor(descriptor_id)

    @jwt_required()
    def put(self, project_id, descriptor_id):
        """
        Update metadata descriptor
        ---
        description: Update a metadata descriptor. Descriptors serve to
          describe extra fields listed in the data attribute of entities.
        tags:
          - Projects
        parameters:
          - in: path
            name: project_id
            required: true
            schema:
              type: string
              format: uuid
            description: Project unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: descriptor_id
            required: true
            schema:
              type: string
              format: uuid
            description: Metadata descriptor unique identifier
            example: b35b7fb5-df86-5776-b181-68564193d36
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: object
                properties:
                  name:
                    type: string
                    description: Name of the metadata descriptor
                    example: "Custom Field"
                  for_client:
                    type: string
                    description: Whether the descriptor is for client
                    default: "False"
                    example: "True"
                  choices:
                    type: array
                    description: List of choices for the descriptor
                    items:
                      type: string
                    example: ["option1", "option2"]
                  departments:
                    type: array
                    description: List of departments for the descriptor
                    items:
                      type: string
                    example: ["department1", "department2"]
        responses:
          200:
            description: Metadata descriptor updated
            content:
              application/json:
                schema:
                  type: object
        """
        args = self.get_args(
            [
                ("name", "", False),
                ("for_client", "False", False),
                ("data_type", "string", True),
                ("choices", [], False, str, "append"),
                ("departments", [], False, str, "append"),
            ]
        )
        user_service.check_all_departments_access(
            project_id,
            projects_service.get_metadata_descriptor(descriptor_id)[
                "departments"
            ]
            + args["departments"],
        )

        if len(args["name"]) == 0:
            raise WrongParameterException("Name cannot be empty.")

        types = [type_name for type_name, _ in METADATA_DESCRIPTOR_TYPES]
        if args["data_type"] not in types:
            raise WrongParameterException("Invalid data_type")

        args["for_client"] = args["for_client"] == "True"

        return projects_service.update_metadata_descriptor(descriptor_id, args)

    @jwt_required()
    def delete(self, project_id, descriptor_id):
        """
        Delete metadata descriptor
        ---
        description: Delete a metadata descriptor. Descriptors serve to describe
          extra fields listed in the data attribute of entities.
        tags:
          - Projects
        parameters:
          - in: path
            name: project_id
            required: true
            schema:
              type: string
              format: uuid
            description: Project unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: descriptor_id
            required: true
            schema:
              type: string
              format: uuid
            description: Metadata descriptor unique identifier
            example: b35b7fb5-df86-5776-b181-68564193d36
        responses:
          204:
            description: Metadata descriptor deleted
        """
        user_service.check_all_departments_access(
            project_id,
            projects_service.get_metadata_descriptor(descriptor_id)[
                "departments"
            ],
        )
        projects_service.remove_metadata_descriptor(descriptor_id)
        return "", 204


class ProductionMetadataDescriptorsReorderResource(Resource, ArgsMixin):

    @jwt_required()
    def post(self, project_id):
        """
        Reorder metadata descriptors
        ---
        description: Reorder metadata descriptors for a specific entity type
          and project. Descriptors are reordered based on the list of
          descriptor IDs provided in the request body. Position is set
          according to the order in the list.
        tags:
          - Projects
        parameters:
          - in: path
            name: project_id
            required: true
            schema:
              type: string
              format: uuid
            description: Project unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: object
                required:
                  - entity_type
                  - descriptor_ids
                properties:
                  entity_type:
                    type: string
                    description: Entity type for the metadata descriptors
                    enum: ["Asset", "Shot", "Edit", "Episode", "Sequence"]
                    example: "Asset"
                  descriptor_ids:
                    type: array
                    description: List of metadata descriptor IDs in the
                      desired order
                    items:
                      type: string
                      format: uuid
                    example: ["b35b7fb5-df86-5776-b181-68564193d36", "c46c8gc6-eg97-6887-c292-79675204e47"]
        responses:
          200:
            description: Metadata descriptors reordered successfully
            content:
              application/json:
                schema:
                  type: array
                  items:
                    type: object
                    properties:
                      id:
                        type: string
                        format: uuid
                        description: Metadata descriptor unique identifier
                        example: a24a6ea4-ce75-4665-a070-57453082c25
                      name:
                        type: string
                        description: Metadata descriptor name
                        example: "Custom Field"
                      position:
                        type: integer
                        description: Position of the descriptor
                        example: 1
                      entity_type:
                        type: string
                        description: Entity type
                        example: "Asset"
          400:
            description: Invalid parameters or descriptor not found
        """
        args = self.get_args(
            [
                ("entity_type", "", True),
                ("descriptor_ids", [], True, str, "append"),
            ]
        )

        user_service.check_manager_project_access(project_id)

        if args["entity_type"] not in [
            "Asset",
            "Shot",
            "Edit",
            "Episode",
            "Sequence",
        ]:
            raise WrongParameterException(
                "Wrong entity type. Please select Asset, Shot, Sequence, Episode or Edit."
            )

        if not args["descriptor_ids"]:
            raise WrongParameterException("descriptor_ids cannot be empty.")

        return projects_service.reorder_metadata_descriptors(
            project_id, args["entity_type"], args["descriptor_ids"]
        )


class ProductionTimeSpentsResource(Resource):
    """
    Resource to retrieve time spents for given production.
    """

    @jwt_required()
    def get(self, project_id):
        """
        Get production time spents
        ---
        description: Retrieve time spents for given production.
        tags:
          - Projects
        parameters:
          - in: path
            name: project_id
            required: true
            schema:
              type: string
              format: uuid
            description: Project unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
          200:
            description: All time spents of given production
            content:
              application/json:
                schema:
                  type: array
                  items:
                    type: object
        """
        user_service.check_project_access(project_id)
        return tasks_service.get_time_spents_for_project(project_id)


class ProductionMilestonesResource(Resource):
    """
    Resource to retrieve milestones for given production.
    """

    @jwt_required()
    def get(self, project_id):
        """
        Get production milestones
        ---
        description: Retrieve milestones for given production.
        tags:
          - Projects
        parameters:
          - in: path
            name: project_id
            required: true
            schema:
              type: string
              format: uuid
            description: Project unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
          200:
            description: All milestones of given production
            content:
              application/json:
                schema:
                  type: array
                  items:
                    type: object
        """
        user_service.check_project_access(project_id)
        return schedule_service.get_milestones_for_project(project_id)


class ProductionScheduleItemsResource(Resource):
    """
    Resource to retrieve schedule items for given production.
    """

    @jwt_required()
    def get(self, project_id):
        """
        Get production schedule items
        ---
        description: Retrieve schedule items for given production.
        tags:
          - Projects
        parameters:
          - in: path
            name: project_id
            required: true
            schema:
              type: string
              format: uuid
            description: Project unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
          200:
            description: All schedule items of given production
            content:
              application/json:
                schema:
                  type: array
                  items:
                    type: object
        """
        user_service.check_project_access(project_id)
        user_service.block_access_to_vendor()
        return schedule_service.get_schedule_items(project_id)


class ProductionTaskTypeScheduleItemsResource(Resource):
    """
    Resource to retrieve schedule items for given production.
    """

    @jwt_required()
    def get(self, project_id):
        """
        Get production task type schedule items
        ---
        description: Retrieve task type schedule items for given production.
        tags:
          - Projects
        parameters:
          - in: path
            name: project_id
            required: true
            schema:
              type: string
              format: uuid
            description: Project unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
          200:
            description: All task types schedule items of given production
            content:
              application/json:
                schema:
                  type: array
                  items:
                    type: object
        """
        user_service.check_project_access(project_id)
        user_service.block_access_to_vendor()
        return schedule_service.get_task_types_schedule_items(project_id)


class ProductionAssetTypesScheduleItemsResource(Resource):
    """
    Resource to retrieve asset types schedule items for given task type.
    """

    @jwt_required()
    def get(self, project_id, task_type_id):
        """
        Get asset types schedule items
        ---
        description: Retrieve asset types schedule items for given task type.
        tags:
          - Projects
        parameters:
          - in: path
            name: project_id
            required: true
            schema:
              type: string
              format: uuid
            description: Project unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: task_type_id
            required: true
            schema:
              type: string
              format: uuid
            description: Task type unique identifier
            example: b35b7fb5-df86-5776-b181-68564193d36
        responses:
          200:
            description: All asset types schedule items for given task type
            content:
              application/json:
                schema:
                  type: array
                  items:
                    type: object
        """
        user_service.check_project_access(project_id)
        user_service.block_access_to_vendor()
        return schedule_service.get_asset_types_schedule_items(
            project_id, task_type_id
        )


class ProductionEpisodesScheduleItemsResource(Resource):
    """
    Resource to retrieve episodes schedule items for given task type.
    """

    @jwt_required()
    def get(self, project_id, task_type_id):
        """
        Get episodes schedule items
        ---
        description: Retrieve episodes schedule items for given task type.
        tags:
          - Projects
        parameters:
          - in: path
            name: project_id
            required: true
            schema:
              type: string
              format: uuid
            description: Project unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: task_type_id
            required: true
            schema:
              type: string
              format: uuid
            description: Task type unique identifier
            example: b35b7fb5-df86-5776-b181-68564193d36
        responses:
          200:
            description: All episodes schedule items for given task type
            content:
              application/json:
                schema:
                  type: array
                  items:
                    type: object
        """
        user_service.check_project_access(project_id)
        user_service.block_access_to_vendor()
        return schedule_service.get_episodes_schedule_items(
            project_id, task_type_id
        )


class ProductionSequencesScheduleItemsResource(Resource):
    """
    Resource to retrieve sequences schedule items for given task type.
    """

    @jwt_required()
    def get(self, project_id, task_type_id):
        """
        Get sequences schedule items
        ---
        description: Retrieve sequences schedule items for given task type.
        tags:
          - Projects
        parameters:
          - in: path
            name: project_id
            required: true
            schema:
              type: string
              format: uuid
            description: Project unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: task_type_id
            required: true
            schema:
              type: string
              format: uuid
            description: Task type unique identifier
            example: b35b7fb5-df86-5776-b181-68564193d36
        responses:
          200:
            description: All sequences schedule items for given task type
            content:
              application/json:
                schema:
                  type: array
                  items:
                    type: object
        """
        user_service.check_project_access(project_id)
        user_service.block_access_to_vendor()
        return schedule_service.get_sequences_schedule_items(
            project_id, task_type_id
        )


class ProductionBudgetsResource(Resource, ArgsMixin):

    @jwt_required()
    def get(self, project_id):
        """
        Get production budgets
        ---
        description: Retrieve budgets for given production.
        tags:
          - Projects
        parameters:
          - in: path
            name: project_id
            required: true
            schema:
              type: string
              format: uuid
            description: Project unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
          200:
            description: All budgets of given production
            content:
              application/json:
                schema:
                  type: array
                  items:
                    type: object
        """
        user_service.check_manager_project_access(project_id)
        self.check_id_parameter(project_id)
        return budget_service.get_budgets(project_id)

    @jwt_required()
    def post(self, project_id):
        """
        Create budget
        ---
        description: Create a budget for given production.
        tags:
          - Projects
        parameters:
          - in: path
            name: project_id
            required: true
            schema:
              type: string
              format: uuid
            description: Project unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: object
                required:
                  - name
                properties:
                  name:
                    type: string
                    description: Budget name
                    example: "New Budget"
                  currency:
                    type: string
                    description: Budget currency code
                    default: "USD"
                    example: "USD"
        responses:
          201:
            description: Budget created
            content:
              application/json:
                schema:
                  type: object
                  properties:
                    id:
                      type: string
                      format: uuid
                      description: Budget unique identifier
                      example: a24a6ea4-ce75-4665-a070-57453082c25
                    name:
                      type: string
                      description: Budget name
                      example: "New Budget"
                    currency:
                      type: string
                      description: Budget currency code
                      example: "USD"
          400:
            description: Invalid parameters
        """
        self.check_id_parameter(project_id)
        user_service.check_manager_project_access(project_id)
        data = self.get_args([("name", None, True), ("currency", None, False)])
        if data["currency"] is None:
            data["currency"] = "USD"
        return budget_service.create_budget(
            project_id, data["name"], data["currency"]
        )


class ProductionBudgetResource(Resource, ArgsMixin):
    """
    Resource to retrieve a budget for given production.
    """

    @jwt_required()
    def get(self, project_id, budget_id):
        """
        Get budget
        ---
        description: Retrieve a budget for given production.
        tags:
          - Projects
        parameters:
          - in: path
            name: project_id
            required: true
            schema:
              type: string
              format: uuid
            description: Project unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: budget_id
            required: true
            schema:
              type: string
              format: uuid
            description: Budget unique identifier
            example: b35b7fb5-df86-5776-b181-68564193d36
        responses:
          200:
            description: Budget retrieved
            content:
              application/json:
                schema:
                  type: object
        """
        self.check_id_parameter(project_id)
        self.check_id_parameter(budget_id)
        user_service.check_manager_project_access(project_id)
        return budget_service.get_budget(budget_id)

    @jwt_required()
    def put(self, project_id, budget_id):
        """
        Update budget
        ---
        description: Update a budget for given production.
        tags:
          - Projects
        parameters:
          - in: path
            name: project_id
            required: true
            schema:
              type: string
              format: uuid
            description: Project unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: budget_id
            required: true
            schema:
              type: string
              format: uuid
            description: Budget unique identifier
            example: b35b7fb5-df86-5776-b181-68564193d36
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: object
                properties:
                  name:
                    type: string
                    description: Budget name
                    example: "New Budget"
                  currency:
                    type: string
                    description: Budget currency code
                    example: "USD"
        responses:
          200:
            description: Budget updated
            content:
              application/json:
                schema:
                  type: object
        """
        self.check_id_parameter(project_id)
        self.check_id_parameter(budget_id)
        user_service.check_manager_project_access(project_id)
        data = self.get_args(
            [("name", None, False), ("currency", None, False)]
        )
        return budget_service.update_budget(
            budget_id, name=data["name"], currency=data["currency"]
        )

    @jwt_required()
    def delete(self, project_id, budget_id):
        """
        Delete budget
        ---
        description: Delete a budget for given production.
        tags:
          - Projects
        parameters:
          - in: path
            name: project_id
            required: true
            schema:
              type: string
              format: uuid
            description: Project unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: budget_id
            required: true
            schema:
              type: string
              format: uuid
            description: Budget unique identifier
            example: b35b7fb5-df86-5776-b181-68564193d36
        responses:
          204:
            description: Budget deleted
        """
        self.check_id_parameter(project_id)
        self.check_id_parameter(budget_id)
        user_service.check_manager_project_access(project_id)
        budget_service.delete_budget(budget_id)
        return "", 204


class ProductionBudgetEntriesResource(Resource, ArgsMixin):

    @jwt_required()
    def get(self, project_id, budget_id):
        """
        Get budget entries
        ---
        description: Retrieve budget entries for given production.
        tags:
          - Projects
        parameters:
          - in: path
            name: project_id
            required: true
            schema:
              type: string
              format: uuid
            description: Project unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: budget_id
            required: true
            schema:
              type: string
              format: uuid
            description: Budget unique identifier
            example: b35b7fb5-df86-5776-b181-68564193d36
        responses:
          200:
            description: All budget entries of given production and budget
            content:
              application/json:
                schema:
                  type: array
                  items:
                    type: object
        """
        self.check_id_parameter(project_id)
        self.check_id_parameter(budget_id)
        user_service.check_manager_project_access(project_id)
        return budget_service.get_budget_entries(budget_id)

    @jwt_required()
    def post(self, project_id, budget_id):
        """
        Create budget entry
        ---
        description: Create a budget entry for given production and budget.
        tags:
          - Projects
        parameters:
          - in: path
            name: project_id
            required: true
            schema:
              type: string
              format: uuid
            description: Project unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: budget_id
            required: true
            schema:
              type: string
              format: uuid
            description: Budget unique identifier
            example: b35b7fb5-df86-5776-b181-68564193d36
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: object
                required:
                  - department_id
                properties:
                  department_id:
                    type: string
                    format: uuid
                    description: Department unique identifier
                    example: c46c8gc6-eg97-6887-c292-79675204e47
                  person_id:
                    type: string
                    format: uuid
                    description: Person unique identifier
                    example: a24a6ea4-ce75-4665-a070-57453082c25
                  start_date:
                    type: string
                    format: date
                    description: Budget entry start date
                    example: "2025-01-01"
                  months_duration:
                    type: integer
                    description: Budget entry duration in months
                    example: 12
                  daily_salary:
                    type: number
                    format: float
                    description: Daily salary amount
                    example: 100.00
                  position:
                    type: string
                    description: Position name
                    example: "Artist"
                  seniority:
                    type: string
                    description: Seniority level
                    example: "Mid"
        responses:
          201:
            description: Budget entry created
            content:
              application/json:
                schema:
                  type: object
                  properties:
                    id:
                      type: string
                      format: uuid
                      description: Budget entry unique identifier
                      example: a24a6ea4-ce75-4665-a070-57453082c25
                    department_id:
                      type: string
                      format: uuid
                      description: Department unique identifier
                      example: c46c8gc6-eg97-6887-c292-79675204e47
          400:
            description: Invalid parameters
        """
        self.check_id_parameter(project_id)
        self.check_id_parameter(budget_id)
        user_service.check_manager_project_access(project_id)
        data = self.get_args(
            [
                ("department_id", None, True),
                ("person_id", None, False),
                ("start_date", None, False),
                ("months_duration", None, False),
                ("daily_salary", None, False),
                ("position", None, False),
                ("seniority", None, False),
            ]
        )
        return budget_service.create_budget_entry(
            budget_id,
            data["department_id"],
            data["start_date"],
            data["months_duration"],
            data["daily_salary"],
            data["position"],
            data["seniority"],
            person_id=data["person_id"],
        )


class ProductionBudgetEntryResource(Resource, ArgsMixin):

    @jwt_required()
    def get(self, project_id, budget_id, entry_id):
        """
        Get budget entry
        ---
        description: Retrieve a budget entry for given production and budget.
        tags:
          - Projects
        parameters:
          - in: path
            name: project_id
            required: true
            schema:
              type: string
              format: uuid
            description: Project unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: budget_id
            required: true
            schema:
              type: string
              format: uuid
            description: Budget unique identifier
            example: b35b7fb5-df86-5776-b181-68564193d36
          - in: path
            name: entry_id
            required: true
            schema:
              type: string
              format: uuid
            description: Budget entry unique identifier
            example: c46c8gc6-eg97-6887-c292-79675204e47
        responses:
          200:
            description: Budget entry retrieved
            content:
              application/json:
                schema:
                  type: object
        """
        user_service.check_manager_project_access(project_id)
        self.check_id_parameter(project_id)
        self.check_id_parameter(budget_id)
        self.check_id_parameter(entry_id)
        return budget_service.get_budget_entry(entry_id)

    @jwt_required()
    def put(self, project_id, budget_id, entry_id):
        """
        Update budget entry
        ---
        description: Update a budget entry for given production and budget.
        tags:
          - Projects
        parameters:
          - in: path
            name: project_id
            required: true
            schema:
              type: string
              format: uuid
            description: Project unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: budget_id
            required: true
            schema:
              type: string
              format: uuid
            description: Budget unique identifier
            example: b35b7fb5-df86-5776-b181-68564193d36
          - in: path
            name: entry_id
            required: true
            schema:
              type: string
              format: uuid
            description: Budget entry unique identifier
            example: c46c8gc6-eg97-6887-c292-79675204e47
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: object
                properties:
                  department_id:
                    type: string
                    format: uuid
                    description: Department unique identifier
                    example: c46c8gc6-eg97-6887-c292-79675204e47
                  person_id:
                    type: string
                    format: uuid
                    description: Person unique identifier
                    example: a24a6ea4-ce75-4665-a070-57453082c25
                  start_date:
                    type: string
                    format: date
                    description: Budget entry start date
                    example: "2025-01-01"
                  months_duration:
                    type: integer
                    description: Budget entry duration in months
                    example: 12
                  daily_salary:
                    type: number
                    format: float
                    description: Daily salary amount
                    example: 100.00
                  position:
                    type: string
                    description: Position name
                    example: "Artist"
                  seniority:
                    type: string
                    description: Seniority level
                    example: "Mid"
                  exceptions:
                    type: object
                    description: Map of amount exceptions. Key is the date and
                      value is the amount
                    example: {"2025-01-01": 1000, "2025-02-01": 2000}
        responses:
          200:
            description: Budget entry updated
            content:
              application/json:
                schema:
                  type: object
        """
        user_service.check_manager_project_access(project_id)
        self.check_id_parameter(project_id)
        self.check_id_parameter(budget_id)
        self.check_id_parameter(entry_id)
        data = self.get_args(
            [
                ("department_id", None, False),
                ("person_id", None, False),
                ("start_date", None, False),
                ("months_duration", None, False),
                ("daily_salary", None, False),
                ("position", None, False),
                ("seniority", None, False),
                {
                    "name": "exceptions",
                    "required": False,
                    "default": {},
                    "type": dict,
                    "help": "Map of amount exceptions. Key is the date and value is the amount. Example: {'2025-01-01': 1000, '2025-02-01': 2000}",
                },
            ]
        )
        return budget_service.update_budget_entry(entry_id, data)

    @jwt_required()
    def delete(self, project_id, budget_id, entry_id):
        """
        Delete budget entry
        ---
        description: Delete a budget entry for given production and budget.
        tags:
          - Projects
        parameters:
          - in: path
            name: project_id
            required: true
            schema:
              type: string
              format: uuid
            description: Project unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: budget_id
            required: true
            schema:
              type: string
              format: uuid
            description: Budget unique identifier
            example: b35b7fb5-df86-5776-b181-68564193d36
          - in: path
            name: entry_id
            required: true
            schema:
              type: string
              format: uuid
            description: Budget entry unique identifier
            example: c46c8gc6-eg97-6887-c292-79675204e47
        responses:
          204:
            description: Budget entry deleted
        """
        self.check_id_parameter(project_id)
        self.check_id_parameter(budget_id)
        self.check_id_parameter(entry_id)
        user_service.check_manager_project_access(project_id)
        budget_service.delete_budget_entry(entry_id)
        return "", 204


class ProductionMonthTimeSpentsResource(Resource, ArgsMixin):

    @jwt_required()
    def get(self, project_id):
        """
        Get production month time spents
        ---
        description: Get aggregated time spents by month for given project.
        tags:
          - Projects
        parameters:
          - in: path
            name: project_id
            required: true
            schema:
              type: string
              format: uuid
            description: Project unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
          200:
            description: Aggregated time spents for given project and month
            content:
              application/json:
                schema:
                  type: object
          400:
            description: Wrong ID format
        """
        permissions.check_admin_permissions()
        self.check_id_parameter(project_id)
        user = persons_service.get_current_user()
        return time_spents_service.get_project_month_time_spents(
            project_id, user["timezone"]
        )


class ProductionScheduleVersionTaskLinksResource(Resource, ArgsMixin):

    @jwt_required()
    def get(self, production_schedule_version_id):
        """
        Get production schedule version task links
        ---
        description: Get task links for given production schedule version.
        tags:
          - Projects
        parameters:
          - in: path
            name: production_schedule_version_id
            required: true
            schema:
              type: string
              format: uuid
            description: Production schedule version unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: query
            name: task_type_id
            required: false
            schema:
              type: string
              format: uuid
            description: Task type unique identifier for filtering
            example: b35b7fb5-df86-5776-b181-68564193d36
        responses:
          200:
            description: Task links for given production schedule version
            content:
              application/json:
                schema:
                  type: array
                  items:
                    type: object
          400:
            description: Wrong ID format
        """
        production_schedule_version = (
            schedule_service.get_production_schedule_version(
                production_schedule_version_id
            )
        )
        if (
            permissions.has_vendor_permissions()
            or permissions.has_client_permissions()
        ):
            raise permissions.PermissionDenied
        user_service.check_project_access(
            production_schedule_version["project_id"]
        )

        args = self.get_args(
            [
                ("task_type_id", None, False),
            ]
        )

        relations = self.get_relations()

        return schedule_service.get_production_schedule_version_task_links(
            production_schedule_version_id,
            task_type_id=args["task_type_id"],
            relations=relations,
        )


class ProductionScheduleVersionSetTaskLinksFromTasksResource(
    Resource, ArgsMixin
):

    @jwt_required()
    def post(self, production_schedule_version_id):
        """
        Set task links from tasks
        ---
        description: Set task links for given production schedule version from
          tasks.
        tags:
          - Projects
        parameters:
          - in: path
            name: production_schedule_version_id
            required: true
            schema:
              type: string
              format: uuid
            description: Production schedule version unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
          200:
            description: Task links created
            content:
              application/json:
                schema:
                  type: object
          400:
            description: Wrong ID format
        """
        production_schedule_version = (
            schedule_service.get_production_schedule_version(
                production_schedule_version_id
            )
        )
        user_service.check_manager_project_access(
            production_schedule_version["project_id"]
        )

        return schedule_service.set_production_schedule_version_task_links_from_production(
            production_schedule_version_id
        )


class ProductionScheduleVersionApplyToProductionResource(Resource, ArgsMixin):

    @jwt_required()
    def post(self, production_schedule_version_id):
        """
        Apply production schedule version
        ---
        description: Apply production schedule version to production.
        tags:
          - Projects
        parameters:
          - in: path
            name: production_schedule_version_id
            required: true
            schema:
              type: string
              format: uuid
            description: Production schedule version unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
          200:
            description: Production schedule version applied
            content:
              application/json:
                schema:
                  type: object
          400:
            description: Wrong ID format
        """
        production_schedule_version = (
            schedule_service.get_production_schedule_version(
                production_schedule_version_id
            )
        )
        user_service.check_manager_project_access(
            production_schedule_version["project_id"]
        )

        return (
            schedule_service.apply_production_schedule_version_to_production(
                production_schedule_version_id,
            )
        )


class ProductionScheduleVersionSetTaskLinksFromProductionScheduleVersionResource(
    Resource, ArgsMixin
):

    @jwt_required()
    def post(self, production_schedule_version_id):
        """
        Set task links from production schedule version
        ---
        description: Set task links for given production schedule version from
          another production schedule version.
        tags:
          - Projects
        parameters:
          - in: path
            name: production_schedule_version_id
            required: true
            schema:
              type: string
              format: uuid
            description: Production schedule version unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: object
                required:
                  - production_schedule_version_id
                properties:
                  production_schedule_version_id:
                    type: string
                    format: uuid
                    description: Source production schedule version unique
                      identifier
                    example: b35b7fb5-df86-5776-b181-68564193d36
        responses:
          200:
            description: Task links created
            content:
              application/json:
                schema:
                  type: object
          400:
            description: Wrong ID format
        """
        production_schedule_version = (
            schedule_service.get_production_schedule_version(
                production_schedule_version_id
            )
        )
        user_service.check_manager_project_access(
            production_schedule_version["project_id"]
        )

        args = self.get_args(
            [
                ("production_schedule_version_id", None, True),
            ]
        )

        other_production_schedule_version = (
            schedule_service.get_production_schedule_version(
                args["production_schedule_version_id"]
            )
        )

        if (
            production_schedule_version["project_id"]
            != other_production_schedule_version["project_id"]
        ):
            raise WrongParameterException(
                "Production schedule versions must belong to the same project."
            )

        return schedule_service.set_production_schedule_version_task_links_from_production_schedule_version(
            production_schedule_version_id,
            other_production_schedule_version_id=other_production_schedule_version[
                "id"
            ],
        )


class ProductionTaskTypesTimeSpentsResource(Resource, ArgsMixin):
    """
    Retrieve time spents for a task type in the production
    """

    @jwt_required()
    def get(self, project_id, task_type_id):
        """
        Get production task type time spents
        ---
        description: Retrieve time spents for a task type in the production.
        tags:
          - Projects
        parameters:
          - in: path
            name: project_id
            required: true
            schema:
              type: string
              format: uuid
            description: Project unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: task_type_id
            required: true
            schema:
              type: string
              format: uuid
            description: Task type unique identifier
            example: b35b7fb5-df86-5776-b181-68564193d36
          - in: query
            name: start_date
            required: false
            schema:
              type: string
              format: date
            description: Start date for filtering time spents
            example: "2022-07-01"
          - in: query
            name: end_date
            required: false
            schema:
              type: string
              format: date
            description: End date for filtering time spents
            example: "2022-07-31"
        responses:
          200:
            description: All time spents for given task type and project
            content:
              application/json:
                schema:
                  type: dict
                  items:
                    type: object
                    properties:
                      id:
                        type: string
                        format: uuid
                        description: Time spent unique identifier
                        example: a24a6ea4-ce75-4665-a070-57453082c25
                      person_id:
                        type: string
                        format: uuid
                        description: Person unique identifier
                        example: b35b7fb5-df86-5776-b181-68564193d36
                      duration:
                        type: number
                        format: float
                        description: Time spent duration in hours
                        example: 8.5
                      date:
                        type: string
                        format: date
                        description: Date of time spent entry
                        example: "2022-07-15"
          400:
            description: Invalid date range parameters
        """
        user_service.check_manager_project_access(project_id)
        arguments = self.get_args(["start_date", "end_date"])
        start_date, end_date = arguments["start_date"], arguments["end_date"]
        try:
            return time_spents_service.get_project_task_type_time_spents(
                project_id, task_type_id, start_date, end_date
            )
        except WrongDateFormatException:
            abort(
                400,
                f"Wrong date format for {start_date} and/or {end_date}",
            )


class ProductionDayOffsResource(Resource, ArgsMixin):
    """
    Retrieve all day offs for a production
    """

    @jwt_required()
    def get(self, project_id):
        """
        Get production day offs
        ---
        description: Retrieve all day offs for a production.
        tags:
          - Projects
        parameters:
          - in: path
            name: project_id
            required: true
            schema:
              type: string
              format: uuid
            description: Project unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: query
            name: start_date
            required: false
            schema:
              type: string
              format: date
            description: Start date for filtering day offs
            example: "2022-07-01"
          - in: query
            name: end_date
            required: false
            schema:
              type: string
              format: date
            description: End date for filtering day offs
            example: "2022-07-31"
        responses:
          200:
            description: All day offs for given project
            content:
              application/json:
                schema:
                  type: dict
                  items:
                    type: object
                    properties:
                      id:
                        type: string
                        format: uuid
                        description: Day off unique identifier
                        example: a24a6ea4-ce75-4665-a070-57453082c25
                      person_id:
                        type: string
                        format: uuid
                        description: Person unique identifier
                        example: b35b7fb5-df86-5776-b181-68564193d36
                      description:
                        type: string
                        description: Day off description
                        example: "Vacation"
                      date:
                        type: string
                        format: date
                        description: Day off start date
                        example: "2022-07-15"
                      end_date:
                        type: string
                        format: date
                        description: Day off end date
                        example: "2022-07-22"
          400:
            description: Invalid date range parameters
        """
        user_service.check_project_access(project_id)
        if (
            permissions.has_client_permissions()
            or permissions.has_vendor_permissions()
        ):
            raise permissions.PermissionDenied
        arguments = self.get_args(["start_date", "end_date"])
        start_date, end_date = arguments["start_date"], arguments["end_date"]
        try:
            return time_spents_service.get_day_offs_between_for_project(
                project_id,
                start_date,
                end_date,
                safe=permissions.has_manager_permissions(),
                current_user_id=persons_service.get_current_user()["id"],
            )
        except WrongDateFormatException:
            abort(
                400,
                f"Wrong date format for {start_date} and/or {end_date}",
            )



from flask import request
from flask_restful import Resource
from flask_jwt_extended import jwt_required

from zou.app.services import (
    breakdown_service,
    deletion_service,
    entities_service,
    persons_service,
    projects_service,
    playlists_service,
    scenes_service,
    shots_service,
    stats_service,
    tasks_service,
    user_service,
)

from zou.app.mixin import ArgsMixin
from zou.app.utils import fields, query, permissions
from zou.app.services.exception import (
    WrongParameterException,
)


class ShotResource(Resource, ArgsMixin):
    @jwt_required()
    def get(self, shot_id):
        """
        Get shot
        ---
        tags:
        - Shots
        description: Get a shot by id. Returns full shot data. Use this to fetch
          a single shot with all fields needed by the UI.
        parameters:
          - in: path
            name: shot_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: Shot found and returned
                content:
                  application/json:
                    schema:
                      type: object
                      properties:
                        id:
                          type: string
                          format: uuid
                          example: a24a6ea4-ce75-4665-a070-57453082c25
                        name:
                          type: string
                          example: SH010
                        project_id:
                          type: string
                          format: uuid
                          example: b24a6ea4-ce75-4665-a070-57453082c25
                        parent_id:
                          type: string
                          format: uuid
                          example: c24a6ea4-ce75-4665-a070-57453082c25
                        nb_frames:
                          type: integer
                          example: 120
                        data:
                          type: object
                          example: {"camera": "camA", "cut_in": 1001}
        """
        shot = shots_service.get_full_shot(shot_id)
        if shot is None:
            shots_service.clear_shot_cache(shot_id)
            shot = shots_service.get_full_shot(shot_id)
        user_service.check_project_access(shot["project_id"])
        user_service.check_entity_access(shot["id"])
        return shot

    @jwt_required()
    def put(self, shot_id):
        """
        Update shot
        ---
        tags:
        - Shots
        description: Update a shot by id. Only mutable fields are allowed. Send a
          JSON body with the fields to change.
        parameters:
          - in: path
            name: shot_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: object
                description: Data to update on the shot
                properties:
                  name:
                    type: string
                    example: SH010 new name
                  description:
                    type: string
                    example: Update description for the shot
                  nb_frames:
                    type: integer
                    example: 24
                  data:
                    type: object
                    example: {"camera": "camA", "cut_in": 1001}
        responses:
            200:
                description: Shot updated
                content:
                  application/json:
                    schema:
                      type: object
                      properties:
                        id:
                          type: string
                          format: uuid
                          example: a24a6ea4-ce75-4665-a070-57453082c25
                        name:
                          type: string
                          example: SH010
                        project_id:
                          type: string
                          format: uuid
                          example: b24a6ea4-ce75-4665-a070-57453082c25
                        parent_id:
                          type: string
                          format: uuid
                          example: c24a6ea4-ce75-4665-a070-57453082c25
                        nb_frames:
                          type: integer
                          example: 120
                        data:
                          type: object
                          example: {"camera": "camA", "cut_in": 1001}
            400:
                description: Invalid body or unsupported fields
        """
        shot = shots_service.get_shot(shot_id)
        user_service.check_manager_project_access(shot["project_id"])
        data = request.json
        if data is None:
            raise WrongParameterException(
                "Data are empty. Please verify that you sent JSON data and"
                " that you set the right headers."
            )
        for field in [
            "id",
            "created_at",
            "updated_at",
            "instance_casting",
            "project_id",
            "entities_in",
            "entities_out",
            "type",
            "shotgun_id",
            "created_by",
        ]:
            data.pop(field, None)

        return shots_service.update_shot(shot_id, data)

    @jwt_required()
    def delete(self, shot_id):
        """
        Delete shot
        ---
        tags:
        - Shots
        description: Delete a shot by id. Requires manager access or ownership of
          the shot.
        parameters:
          - in: path
            name: shot_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            204:
                description: Shot deleted
        """
        force = self.get_force()
        shot = shots_service.get_shot(shot_id)
        if shot["created_by"] == persons_service.get_current_user()["id"]:
            user_service.check_belong_to_project(shot["project_id"])
        else:
            user_service.check_manager_project_access(shot["project_id"])
        shots_service.remove_shot(shot_id, force=force)
        return "", 204


class SceneResource(Resource):
    @jwt_required()
    def get(self, scene_id):
        """
        Get scene
        ---
        tags:
        - Shots
        description: Get a scene by id. Returns full scene data needed by clients.
        parameters:
          - in: path
            name: scene_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: Scene found and returned
                content:
                  application/json:
                    schema:
                      type: object
                      properties:
                        id:
                          type: string
                          format: uuid
                          example: a24a6ea4-ce75-4665-a070-57453082c25
                        name:
                          type: string
                          example: SC001
                        project_id:
                          type: string
                          format: uuid
                          example: b24a6ea4-ce75-4665-a070-57453082c25
        """
        scene = shots_service.get_full_scene(scene_id)
        user_service.check_project_access(scene["project_id"])
        user_service.check_entity_access(scene["id"])
        return scene

    @jwt_required()
    def delete(self, scene_id):
        """
        Delete scene
        ---
        tags:
        - Shots
        description: Delete a scene by id. Requires manager access or ownership.
        parameters:
          - in: path
            name: scene_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            204:
                description: Scene deleted
        """
        scene = shots_service.get_scene(scene_id)
        if scene["created_by"] == persons_service.get_current_user()["id"]:
            user_service.check_belong_to_project(scene["project_id"])
        else:
            user_service.check_manager_project_access(scene["project_id"])
        shots_service.remove_scene(scene_id)
        return "", 204


class ShotsResource(Resource):
    @jwt_required()
    def get(self):
        """
        Get shots
        ---
        tags:
        - Shots
        description: Get shots with optional filters. Use query params like
          project_id, sequence_id or parent_id to filter results.
        parameters:
          - in: query
            name: sequence_id
            required: False
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: query
            name: project_id
            required: False
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: query
            name: parent_id
            required: False
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: All shot entries
                content:
                  application/json:
                    schema:
                      type: array
                      items:
                        type: object
                        properties:
                          id:
                            type: string
                            format: uuid
                            example: a24a6ea4-ce75-4665-a070-57453082c25
                          name:
                            type: string
                            example: SH010
                          project_id:
                            type: string
                            format: uuid
                            example: b24a6ea4-ce75-4665-a070-57453082c25
                          parent_id:
                            type: string
                            format: uuid
                            example: c24a6ea4-ce75-4665-a070-57453082c25
        """
        criterions = query.get_query_criterions_from_request(request)
        if "sequence_id" in criterions:
            sequence = shots_service.get_sequence(criterions["sequence_id"])
            criterions["project_id"] = sequence["project_id"]
            criterions["parent_id"] = sequence["id"]
            del criterions["sequence_id"]
        user_service.check_project_access(criterions.get("project_id", None))
        if permissions.has_vendor_permissions():
            criterions["assigned_to"] = persons_service.get_current_user()[
                "id"
            ]
        return shots_service.get_shots(criterions)


class AllShotsResource(Resource):
    @jwt_required()
    def get(self):
        """
        Get all shots
        ---
        tags:
        - Shots
        description: Get all shots across projects with optional filters. Use
          sequence_id, project_id, or parent_id to filter.
        parameters:
          - in: query
            name: sequence_id
            required: False
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: query
            name: project_id
            required: False
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: query
            name: parent_id
            required: False
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: All shot entries
                content:
                  application/json:
                    schema:
                      type: array
                      items:
                        type: object
                        properties:
                          id:
                            type: string
                            format: uuid
                            example: a24a6ea4-ce75-4665-a070-57453082c25
                          name:
                            type: string
                            example: SH020
                          project_id:
                            type: string
                            format: uuid
                            example: b24a6ea4-ce75-4665-a070-57453082c25
                          parent_id:
                            type: string
                            format: uuid
                            example: c24a6ea4-ce75-4665-a070-57453082c25
        """
        criterions = query.get_query_criterions_from_request(request)
        if "sequence_id" in criterions:
            sequence = shots_service.get_sequence(criterions["sequence_id"])
            criterions["project_id"] = sequence["project_id"]
            criterions["parent_id"] = sequence["id"]
            del criterions["sequence_id"]
        if permissions.has_vendor_permissions():
            criterions["assigned_to"] = persons_service.get_current_user()[
                "id"
            ]
        user_service.check_project_access(criterions.get("project_id", None))
        return shots_service.get_shots(criterions)


class ScenesResource(Resource):
    @jwt_required()
    def get(self):
        """
        Get scenes
        ---
        tags:
        - Shots
        description: Get scenes with optional filters. Use project_id to filter
          by project.
        parameters:
          - in: query
            name: project_id
            required: False
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: All scene entries
                content:
                  application/json:
                    schema:
                      type: array
                      items:
                        type: object
                        properties:
                          id:
                            type: string
                            format: uuid
                            example: a24a6ea4-ce75-4665-a070-57453082c25
                          name:
                            type: string
                            example: SC001
                          project_id:
                            type: string
                            format: uuid
                            example: b24a6ea4-ce75-4665-a070-57453082c25
        """
        criterions = query.get_query_criterions_from_request(request)
        user_service.check_project_access(criterions.get("project_id", None))
        return shots_service.get_scenes(criterions)


class ShotAssetsResource(Resource):
    @jwt_required()
    def get(self, shot_id):
        """
        Get shot assets
        ---
        tags:
        - Shots
        description: Get assets linked to a shot. Returns the breakdown casting
          for the shot.
        parameters:
          - in: path
            name: shot_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: All assets for given shot
                content:
                  application/json:
                    schema:
                      type: array
                      items:
                        type: object
                        properties:
                          id:
                            type: string
                            format: uuid
                            example: a24a6ea4-ce75-4665-a070-57453082c25
                          name:
                            type: string
                            example: Character A
                          entity_type_id:
                            type: string
                            format: uuid
                            example: b24a6ea4-ce75-4665-a070-57453082c25
        """
        shot = shots_service.get_shot(shot_id)
        user_service.check_project_access(shot["project_id"])
        user_service.check_entity_access(shot["id"])
        return breakdown_service.get_entity_casting(shot_id)


class ShotTaskTypesResource(Resource):
    @jwt_required()
    def get(self, shot_id):
        """
        Get shot task types
        ---
        tags:
        - Shots
        description: Get task types for a shot.
        parameters:
          - in: path
            name: shot_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: All task types related to given shot
                content:
                  application/json:
                    schema:
                      type: array
                      items:
                        type: object
                        properties:
                          id:
                            type: string
                            format: uuid
                            example: a24a6ea4-ce75-4665-a070-57453082c25
                          name:
                            type: string
                            example: Animation
        """
        shot = shots_service.get_shot(shot_id)
        user_service.check_project_access(shot["project_id"])
        user_service.check_entity_access(shot["id"])
        return tasks_service.get_task_types_for_shot(shot_id)


class ShotTasksResource(Resource, ArgsMixin):
    @jwt_required()
    def get(self, shot_id):
        """
        Get shot tasks
        ---
        tags:
        - Shots
        description: Get tasks for a shot. Optionally include relations using
          query params.
        parameters:
          - in: path
            name: shot_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: All tasks related to given shot
                content:
                  application/json:
                    schema:
                      type: array
                      items:
                        type: object
                        properties:
                          id:
                            type: string
                            format: uuid
                            example: a24a6ea4-ce75-4665-a070-57453082c25
                          name:
                            type: string
                            example: SH010 Animation
                          task_type_id:
                            type: string
                            format: uuid
                            example: b24a6ea4-ce75-4665-a070-57453082c25
                          task_status_id:
                            type: string
                            format: uuid
                            example: c24a6ea4-ce75-4665-a070-57453082c25
                          entity_id:
                            type: string
                            format: uuid
                            example: d24a6ea4-ce75-4665-a070-57453082c25
                          project_id:
                            type: string
                            format: uuid
                            example: e24a6ea4-ce75-4665-a070-57453082c25
                          assignees:
                            type: array
                            items:
                              type: string
                              format: uuid
                            example: ["f24a6ea4-ce75-4665-a070-57453082c25"]
        """
        shot = shots_service.get_shot(shot_id)
        user_service.check_project_access(shot["project_id"])
        user_service.check_entity_access(shot["id"])
        relations = self.get_relations()
        return tasks_service.get_tasks_for_shot(shot_id, relations=relations)


class SequenceShotTasksResource(Resource, ArgsMixin):
    @jwt_required()
    def get(self, sequence_id):
        """
        Get sequence shot tasks
        ---
        tags:
        - Shots
        description: Get shot tasks for a sequence. Restricted for vendor
          permissions.
        parameters:
          - in: path
            name: sequence_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: All shot tasks related to given sequence
                content:
                  application/json:
                    schema:
                      type: array
                      items:
                        type: object
                        properties:
                          id:
                            type: string
                            format: uuid
                            example: a24a6ea4-ce75-4665-a070-57453082c25
                          name:
                            type: string
                            example: SH010 Animation
                          task_type_id:
                            type: string
                            format: uuid
                            example: b24a6ea4-ce75-4665-a070-57453082c25
                          entity_id:
                            type: string
                            format: uuid
                            example: c24a6ea4-ce75-4665-a070-57453082c25
                          task_status_id:
                            type: string
                            format: uuid
                            example: d24a6ea4-ce75-4665-a070-57453082c25
        """
        sequence = shots_service.get_sequence(sequence_id)
        user_service.check_project_access(sequence["project_id"])
        user_service.check_entity_access(sequence["id"])
        if permissions.has_vendor_permissions():
            raise permissions.PermissionDenied
        relations = self.get_relations()
        return tasks_service.get_shot_tasks_for_sequence(
            sequence_id, relations=relations
        )


class EpisodeShotTasksResource(Resource, ArgsMixin):
    @jwt_required()
    def get(self, episode_id):
        """
        Get episode shot tasks
        ---
        tags:
        - Shots
        description: Get shot tasks for an episode. Restricted for vendor
          permissions.
        parameters:
          - in: path
            name: episode_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: All shot tasks related to given episode
                content:
                  application/json:
                    schema:
                      type: array
                      items:
                        type: object
                        properties:
                          id:
                            type: string
                            format: uuid
                            example: a24a6ea4-ce75-4665-a070-57453082c25
                          name:
                            type: string
                            example: EP01 Layout
                          task_type_id:
                            type: string
                            format: uuid
                            example: b24a6ea4-ce75-4665-a070-57453082c25
                          entity_id:
                            type: string
                            format: uuid
                            example: c24a6ea4-ce75-4665-a070-57453082c25
                          task_status_id:
                            type: string
                            format: uuid
                            example: d24a6ea4-ce75-4665-a070-57453082c25
        """
        episode = shots_service.get_episode(episode_id)
        user_service.check_project_access(episode["project_id"])
        user_service.check_entity_access(episode["id"])
        if permissions.has_vendor_permissions():
            raise permissions.PermissionDenied
        relations = self.get_relations()
        return tasks_service.get_shot_tasks_for_episode(
            episode_id, relations=relations
        )


class EpisodeAssetTasksResource(Resource, ArgsMixin):
    @jwt_required()
    def get(self, episode_id):
        """
        Get episode asset tasks
        ---
        tags:
        - Shots
        description: Get asset tasks for an episode. Restricted for vendor
          permissions.
        parameters:
          - in: path
            name: episode_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: All assets tasks related to given episode
                content:
                  application/json:
                    schema:
                      type: array
                      items:
                        type: object
                        properties:
                          id:
                            type: string
                            format: uuid
                            example: a24a6ea4-ce75-4665-a070-57453082c25
                          name:
                            type: string
                            example: Character Modeling
                          task_type_id:
                            type: string
                            format: uuid
                            example: b24a6ea4-ce75-4665-a070-57453082c25
                          entity_id:
                            type: string
                            format: uuid
                            example: c24a6ea4-ce75-4665-a070-57453082c25
                          task_status_id:
                            type: string
                            format: uuid
                            example: d24a6ea4-ce75-4665-a070-57453082c25
        """
        episode = shots_service.get_episode(episode_id)
        user_service.check_project_access(episode["project_id"])
        user_service.check_entity_access(episode["id"])
        if permissions.has_vendor_permissions():
            raise permissions.PermissionDenied
        relations = self.get_relations()
        return tasks_service.get_asset_tasks_for_episode(
            episode_id, relations=relations
        )


class EpisodeShotsResource(Resource, ArgsMixin):
    @jwt_required()
    def get(self, episode_id):
        """
        Get episode shots
        ---
        tags:
        - Shots
        description: Get shots for an episode. Supports including relations via
          query params.
        parameters:
          - in: path
            name: episode_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: All shots related to given episode
                content:
                  application/json:
                    schema:
                      type: array
                      items:
                        type: object
                        properties:
                          id:
                            type: string
                            format: uuid
                            example: a24a6ea4-ce75-4665-a070-57453082c25
                          name:
                            type: string
                            example: SH010
                          project_id:
                            type: string
                            format: uuid
                            example: b24a6ea4-ce75-4665-a070-57453082c25
                          parent_id:
                            type: string
                            format: uuid
                            example: c24a6ea4-ce75-4665-a070-57453082c25
        """
        episode = shots_service.get_episode(episode_id)
        user_service.check_project_access(episode["project_id"])
        user_service.check_entity_access(episode["id"])
        relations = self.get_relations()
        return shots_service.get_shots_for_episode(
            episode_id, relations=relations
        )


class ShotPreviewsResource(Resource):
    @jwt_required()
    def get(self, shot_id):
        """
        Get shot previews
        ---
        tags:
        - Shots
        description: Return previews for a shot as a dict keyed by task type id.
          Each value is an array of previews for that task type.
        parameters:
          - in: path
            name: shot_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: All previews related to given shot
                content:
                  application/json:
                    schema:
                      type: object
                      additionalProperties:
                        type: array
                        items:
                          type: object
                          properties:
                            id:
                              type: string
                              format: uuid
                              example: b24a6ea4-ce75-4665-a070-57453082c25
                            revision:
                              type: integer
                              example: 3
                            file_id:
                              type: string
                              format: uuid
                              example: c24a6ea4-ce75-4665-a070-57453082c25
                    example:
                      "a24a6ea4-ce75-4665-a070-57453082c25": [
                        {
                          "id": "b24a6ea4-ce75-4665-a070-57453082c25",
                          "revision": 3,
                          "file_id": "c24a6ea4-ce75-4665-a070-57453082c25"
                        }
                      ]
        """
        shot = shots_service.get_shot(shot_id)
        user_service.check_project_access(shot["project_id"])
        user_service.check_entity_access(shot["id"])
        return playlists_service.get_preview_files_for_entity(shot_id)


class SequenceTasksResource(Resource, ArgsMixin):
    @jwt_required()
    def get(self, sequence_id):
        """
        Get sequence tasks
        ---
        tags:
        - Shots
        description: Get tasks for a sequence. Optionally include relations using
          query params.
        parameters:
          - in: path
            name: sequence_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: All tasks related to given shot
                content:
                  application/json:
                    schema:
                      type: array
                      items:
                        type: object
                        properties:
                          id:
                            type: string
                            format: uuid
                            example: a24a6ea4-ce75-4665-a070-57453082c25
                          name:
                            type: string
                            example: SH010 Animation
                          task_type_id:
                            type: string
                            format: uuid
                            example: b24a6ea4-ce75-4665-a070-57453082c25
                          task_status_id:
                            type: string
                            format: uuid
                            example: c24a6ea4-ce75-4665-a070-57453082c25
                          entity_id:
                            type: string
                            format: uuid
                            example: d24a6ea4-ce75-4665-a070-57453082c25
                          project_id:
                            type: string
                            format: uuid
                            example: e24a6ea4-ce75-4665-a070-57453082c25
        """
        sequence = shots_service.get_sequence(sequence_id)
        user_service.check_project_access(sequence["project_id"])
        relations = self.get_relations()
        return tasks_service.get_tasks_for_sequence(
            sequence_id, relations=relations
        )


class SequenceTaskTypesResource(Resource):
    @jwt_required()
    def get(self, sequence_id):
        """
        Get sequence task types
        ---
        tags:
        - Shots
        description: Get task types for a sequence.
        parameters:
          - in: path
            name: sequence_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: All task types related to given shot
                content:
                  application/json:
                    schema:
                      type: array
                      items:
                        type: object
                        properties:
                          id:
                            type: string
                            format: uuid
                            example: a24a6ea4-ce75-4665-a070-57453082c25
                          name:
                            type: string
                            example: Animation
        """
        sequence = shots_service.get_sequence(sequence_id)
        user_service.check_project_access(sequence["project_id"])
        return tasks_service.get_task_types_for_sequence(sequence_id)


class ShotsAndTasksResource(Resource):

    @jwt_required()
    def get(self):
        """
        Get shots and tasks
        ---
        tags:
        - Shots
        description: Get shots and their related tasks. Optionally filter by project.
        parameters:
          - in: query
            name: project_id
            required: False
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: All shots
                content:
                  application/json:
                    schema:
                      type: array
                      items:
                        type: object
                        properties:
                          id:
                            type: string
                            format: uuid
                            example: a24a6ea4-ce75-4665-a070-57453082c25
                          name:
                            type: string
                            example: SH010
                          project_id:
                            type: string
                            format: uuid
                            example: b24a6ea4-ce75-4665-a070-57453082c25
                          tasks:
                            type: array
                            items:
                              type: object
                              properties:
                                id:
                                  type: string
                                  format: uuid
                                  example: c24a6ea4-ce75-4665-a070-57453082c25
                                name:
                                  type: string
                                  example: SH010 Animation
        """
        criterions = query.get_query_criterions_from_request(request)
        user_service.check_project_access(criterions.get("project_id", None))
        if permissions.has_vendor_permissions():
            criterions["assigned_to"] = persons_service.get_current_user()[
                "id"
            ]
            criterions["vendor_departments"] = [
                str(department.id)
                for department in persons_service.get_current_user_raw().departments
            ]
        return shots_service.get_shots_and_tasks(criterions)


class SceneAndTasksResource(Resource):
    @jwt_required()
    def get(self):
        """
        Get scenes and tasks
        ---
        tags:
        - Shots
        description: Get scenes and their related tasks. Optionally filter by project.
        parameters:
          - in: query
            name: project_id
            required: False
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: All scenes
                content:
                  application/json:
                    schema:
                      type: array
                      items:
                        type: object
                        properties:
                          id:
                            type: string
                            format: uuid
                            example: a24a6ea4-ce75-4665-a070-57453082c25
                          name:
                            type: string
                            example: SC001
                          project_id:
                            type: string
                            format: uuid
                            example: b24a6ea4-ce75-4665-a070-57453082c25
                          tasks:
                            type: array
                            items:
                              type: object
                              properties:
                                id:
                                  type: string
                                  format: uuid
                                  example: c24a6ea4-ce75-4665-a070-57453082c25
                                name:
                                  type: string
                                  example: Layout
        """
        criterions = query.get_query_criterions_from_request(request)
        user_service.check_project_access(criterions.get("project_id", None))
        criterions["entity_type_id"] = shots_service.get_scene_type()["id"]
        return entities_service.get_entities_and_tasks(criterions)


class SequenceAndTasksResource(Resource):
    @jwt_required()
    def get(self):
        """
        Get sequences and tasks
        ---
        tags:
        - Shots
        description: Get sequences and their related tasks.
          Optionally filter by project.
        parameters:
          - in: query
            name: project_id
            required: False
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: All sequences
                content:
                  application/json:
                    schema:
                      type: array
                      items:
                        type: object
                        properties:
                          id:
                            type: string
                            format: uuid
                            example: a24a6ea4-ce75-4665-a070-57453082c25
                          name:
                            type: string
                            example: SQ010
                          project_id:
                            type: string
                            format: uuid
                            example: b24a6ea4-ce75-4665-a070-57453082c25
                          tasks:
                            type: array
                            items:
                              type: object
                              properties:
                                id:
                                  type: string
                                  format: uuid
                                  example: c24a6ea4-ce75-4665-a070-57453082c25
                                name:
                                  type: string
                                  example: SQ010 Editing
        """
        criterions = query.get_query_criterions_from_request(request)
        user_service.check_project_access(criterions.get("project_id", None))
        criterions["entity_type_id"] = shots_service.get_sequence_type()["id"]
        return entities_service.get_entities_and_tasks(criterions)


class EpisodeAndTasksResource(Resource):
    @jwt_required()
    def get(self):
        """
        Get episodes and tasks
        ---
        tags:
        - Shots
        description: Get episodes and their related tasks.
          Optionally filter by project.
        parameters:
          - in: query
            name: project_id
            required: False
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: All episodes
                content:
                  application/json:
                    schema:
                      type: array
                      items:
                        type: object
                        properties:
                          id:
                            type: string
                            format: uuid
                            example: a24a6ea4-ce75-4665-a070-57453082c25
                          name:
                            type: string
                            example: EP01
                          project_id:
                            type: string
                            format: uuid
                            example: b24a6ea4-ce75-4665-a070-57453082c25
                          tasks:
                            type: array
                            items:
                              type: object
                              properties:
                                id:
                                  type: string
                                  format: uuid
                                  example: c24a6ea4-ce75-4665-a070-57453082c25
                                name:
                                  type: string
                                  example: EP01 Layout
        """
        criterions = query.get_query_criterions_from_request(request)
        user_service.check_project_access(criterions.get("project_id", None))
        criterions["entity_type_id"] = shots_service.get_episode_type()["id"]
        return entities_service.get_entities_and_tasks(criterions)


class ProjectShotsResource(Resource, ArgsMixin):
    @jwt_required()
    def get(self, project_id):
        """
        Get project shots
        ---
        tags:
        - Shots
        description: Get shots for a project. May limit to assigned shots for
          vendor users.
        parameters:
          - in: path
            name: project_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: All shots related to given project
                content:
                  application/json:
                    schema:
                      type: array
                      items:
                        type: object
                        properties:
                          id:
                            type: string
                            format: uuid
                            example: a24a6ea4-ce75-4665-a070-57453082c25
                          name:
                            type: string
                            example: SH010
                          project_id:
                            type: string
                            format: uuid
                            example: b24a6ea4-ce75-4665-a070-57453082c25
                          parent_id:
                            type: string
                            format: uuid
                            example: c24a6ea4-ce75-4665-a070-57453082c25
        """
        projects_service.get_project(project_id)
        user_service.check_project_access(project_id)
        return shots_service.get_shots_for_project(
            project_id, only_assigned=permissions.has_vendor_permissions()
        )

    @jwt_required()
    def post(self, project_id):
        """
        Create project shot
        ---
        tags:
        - Shots
        description: Create a shot in a project. Provide name and optional
          fields like description, sequence_id and nb_frames.
        parameters:
          - in: path
            name: project_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: object
                required:
                  - name
                properties:
                  name:
                    type: string
                    required: true
                    example: SH010
                  description:
                    type: string
                    required: false
                    example: A short description of the shot
                  sequence_id:
                    type: string
                    format: uuid
                    required: false
                    example: a24a6ea4-ce75-4665-a070-57453082c25
                  nb_frames:
                    type: integer
                    required: false
                    example: 24
        responses:
            201:
                description: Shot created
                content:
                  application/json:
                    schema:
                      type: object
                      properties:
                        id:
                          type: string
                          format: uuid
                          example: a24a6ea4-ce75-4665-a070-57453082c25
                        name:
                          type: string
                          example: SH010
                        project_id:
                          type: string
                          format: uuid
                          example: b24a6ea4-ce75-4665-a070-57453082c25
                        sequence_id:
                          type: string
                          format: uuid
                          example: c24a6ea4-ce75-4665-a070-57453082c25
                        nb_frames:
                          type: integer
                          example: 24
                        description:
                          type: string
                          example: A short description of the shot
        """
        (
            sequence_id,
            name,
            data,
            nb_frames,
            description,
        ) = self.get_arguments()
        projects_service.get_project(project_id)
        user_service.check_manager_project_access(project_id)

        shot = shots_service.create_shot(
            project_id,
            sequence_id,
            name,
            data=data,
            nb_frames=nb_frames,
            description=description,
            created_by=persons_service.get_current_user()["id"],
        )
        return shot, 201

    def get_arguments(self):
        args = self.get_args(
            [
                {"name": "name", "required": True},
                "sequence_id",
                {"name": "data", "type": dict},
                {"name": "nb_frames", "type": int},
                "description",
            ]
        )

        return (
            args["sequence_id"],
            args["name"],
            args["data"],
            args["nb_frames"],
            args["description"],
        )


class ProjectSequencesResource(Resource, ArgsMixin):
    @jwt_required()
    def get(self, project_id):
        """
        Get project sequences
        ---
        tags:
        - Shots
        description: Get sequences for a project. May limit to assigned items for
          vendor users.
        parameters:
          - in: path
            name: project_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: All sequences related to given project
                content:
                  application/json:
                    schema:
                      type: array
                      items:
                        type: object
                        properties:
                          id:
                            type: string
                            format: uuid
                            example: a24a6ea4-ce75-4665-a070-57453082c25
                          name:
                            type: string
                            example: SQ01
                          project_id:
                            type: string
                            format: uuid
                            example: b24a6ea4-ce75-4665-a070-57453082c25
                          episode_id:
                            type: string
                            format: uuid
                            example: c24a6ea4-ce75-4665-a070-57453082c25
        """
        projects_service.get_project(project_id)
        user_service.check_project_access(project_id)
        return shots_service.get_sequences_for_project(
            project_id, only_assigned=permissions.has_vendor_permissions()
        )

    @jwt_required()
    def post(self, project_id):
        """
        Create project sequence
        ---
        tags:
        - Shots
        description: Create a sequence in a project. Provide name and optional
          episode_id.
        parameters:
          - in: path
            name: project_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: object
                required:
                  - name
                properties:
                  name:
                    type: string
                    required: true
                    example: SQ01
                  episode_id:
                    type: string
                    format: uuid
                    required: false
                    example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            201:
                description: Sequence created
                content:
                  application/json:
                    schema:
                      type: object
                      properties:
                        id:
                          type: string
                          format: uuid
                          example: a24a6ea4-ce75-4665-a070-57453082c25
                        name:
                          type: string
                          example: SQ01
                        project_id:
                          type: string
                          format: uuid
                          example: b24a6ea4-ce75-4665-a070-57453082c25
                        episode_id:
                          type: string
                          format: uuid
                          example: c24a6ea4-ce75-4665-a070-57453082c25
                        description:
                          type: string
                          example: A sequence description
        """
        (episode_id, name, description, data) = self.get_arguments()
        projects_service.get_project(project_id)
        user_service.check_manager_project_access(project_id)
        sequence = shots_service.create_sequence(
            project_id,
            episode_id,
            name,
            description=description,
            data=data,
            created_by=persons_service.get_current_user()["id"],
        )
        return sequence, 201

    def get_arguments(self):
        args = self.get_args(
            [
                {"name": "name", "required": True},
                "episode_id",
                {"name": "description", "default": ""},
                {"name": "data", "type": dict, "default": {}},
            ]
        )

        return (
            args["episode_id"],
            args["name"],
            args["description"],
            args["data"],
        )


class ProjectEpisodesResource(Resource, ArgsMixin):
    @jwt_required()
    def get(self, project_id):
        """
        Get project episodes
        ---
        tags:
        - Shots
        description: Get episodes for a project. May limit to assigned items for
          vendor users.
        parameters:
          - in: path
            name: project_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: All episodes related to given project
                content:
                  application/json:
                    schema:
                      type: array
                      items:
                        type: object
                        properties:
                          id:
                            type: string
                            format: uuid
                            example: a24a6ea4-ce75-4665-a070-57453082c25
                          name:
                            type: string
                            example: EP01
                          project_id:
                            type: string
                            format: uuid
                            example: b24a6ea4-ce75-4665-a070-57453082c25
                          description:
                            type: string
                            example: A short description of the episode
                          status:
                            type: string
                            example: running
        """
        projects_service.get_project(project_id)
        user_service.check_project_access(project_id)
        return shots_service.get_episodes_for_project(
            project_id, only_assigned=permissions.has_vendor_permissions()
        )

    @jwt_required()
    def post(self, project_id):
        """
        Create project episode
        ---
        tags:
        - Shots
        description: Create an episode in a project. Provide name and
          description. Status is optional.
        parameters:
          - in: path
            name: project_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: object
                required:
                  - name
                  - description
                properties:
                  name:
                    type: string
                    required: true
                    example: EP01
                  description:
                    type: string
                    required: true
                    example: A short description of the episode
                  status:
                    type: string
                    required: false
                    example: running
        responses:
            201:
                description: Episode created
                content:
                  application/json:
                    schema:
                      type: object
                      properties:
                        id:
                          type: string
                          format: uuid
                          example: a24a6ea4-ce75-4665-a070-57453082c25
                        name:
                          type: string
                          example: EP01
                        project_id:
                          type: string
                          format: uuid
                          example: b24a6ea4-ce75-4665-a070-57453082c25
                        description:
                          type: string
                          example: A short description of the episode
                        status:
                          type: string
                          example: running
        """
        name, status, description, data = self.get_arguments()
        projects_service.get_project(project_id)
        user_service.check_manager_project_access(project_id)
        return (
            shots_service.create_episode(
                project_id,
                name,
                status,
                description,
                data,
                created_by=persons_service.get_current_user()["id"],
            ),
            201,
        )

    def get_arguments(self):
        args = self.get_args(
            [
                {"name": "name", "required": True},
                {"name": "status", "default": "running"},
                {"name": "description", "default": ""},
                {"name": "data", "type": dict, "default": {}},
            ]
        )

        return args["name"], args["status"], args["description"], args["data"]


class ProjectEpisodeStatsResource(Resource):
    @jwt_required()
    def get(self, project_id):
        """
        Get episode stats
        ---
        tags:
        - Shots
        description: Return number of tasks by status, task type and episode
          for the project.
        parameters:
          - in: path
            name: project_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: Number of tasks by status, task types and episodes for given project
                content:
                  application/json:
                    schema:
                      type: object
                      additionalProperties:
                        type: object
                        additionalProperties:
                          type: object
                          additionalProperties:
                            type: object
                            properties:
                              count:
                                type: integer
                                example: 120
                              frames:
                                type: integer
                                example: 1440
                              drawings:
                                type: integer
                                example: 360
                      example:
                        episodeId1:
                          taskTypeId1:
                            taskStatusId1:
                              count: 50
                              frames: 600
                              drawings: 150
                            taskStatusId2:
                              count: 70
                              frames: 840
                              drawings: 210
                        all:
                          all:
                            taskStatusId1:
                              count: 200
                              frames: 2400
                              drawings: 600
        """
        projects_service.get_project(project_id)
        user_service.check_project_access(project_id)
        return stats_service.get_episode_stats_for_project(
            project_id, only_assigned=permissions.has_vendor_permissions()
        )


class ProjectEpisodeRetakeStatsResource(Resource):
    @jwt_required()
    def get(self, project_id):
        """
        Get episode retake stats
        ---
        tags:
        - Shots
        description: Return retake and done counts by task type and episode.
          Includes evolution data and max retake count.
        parameters:
          - in: path
            name: project_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: Number of tasks by status, task types and episodes for given project
                content:
                  application/json:
                    schema:
                      type: object
                      additionalProperties:
                        type: object
                        properties:
                          max_retake_count:
                            type: integer
                            example: 4
                          evolution:
                            type: object
                            additionalProperties:
                              type: object
                              properties:
                                retake:
                                  type: object
                                  properties:
                                    count:
                                      type: integer
                                      example: 80
                                    frames:
                                      type: integer
                                      example: 7900
                                    drawings:
                                      type: integer
                                      example: 8000
                                done:
                                  type: object
                                  properties:
                                    count:
                                      type: integer
                                      example: 117
                                    frames:
                                      type: integer
                                      example: 3900
                                    drawings:
                                      type: integer
                                      example: 8000
                          done:
                            type: object
                            properties:
                              count:
                                type: integer
                                example: 197
                              frames:
                                type: integer
                                example: 16090
                              drawings:
                                type: integer
                                example: 16090
                          retake:
                            type: object
                            properties:
                              count:
                                type: integer
                                example: 0
                              frames:
                                type: integer
                                example: 0
                              drawings:
                                type: integer
                                example: 0
                          other:
                            type: object
                            properties:
                              count:
                                type: integer
                                example: 5
                              frames:
                                type: integer
                                example: 185
                              drawings:
                                type: integer
                                example: 185
                      example:
                        episodeId1:
                          max_retake_count: 4
                          evolution:
                            "1":
                              retake:
                                count: 80
                                frames: 7900
                                drawings: 8000
                              done:
                                count: 117
                                frames: 3900
                                drawings: 8000
                          done:
                            count: 197
                            frames: 16090
                            drawings: 16090
                          retake:
                            count: 0
                            frames: 0
                            drawings: 0
                          other:
                            count: 5
                            frames: 185
                            drawings: 185
                        all:
                          all:
                            max_retake_count: 4
                            done:
                              count: 500
                              frames: 50000
                              drawings: 50000
                            retake:
                              count: 100
                              frames: 10000
                              drawings: 10000
                            other:
                              count: 10
                              frames: 1000
                              drawings: 1000
        """
        projects_service.get_project(project_id)
        user_service.check_project_access(project_id)
        return stats_service.get_episode_retake_stats_for_project(
            project_id, only_assigned=permissions.has_vendor_permissions()
        )


class EpisodeResource(Resource, ArgsMixin):
    @jwt_required()
    def get(self, episode_id):
        """
        Get episode
        ---
        tags:
        - Shots
        description: Get an episode by id.
          needs.
        parameters:
          - in: path
            name: episode_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: Episode found and returned
                content:
                  application/json:
                    schema:
                      type: object
                      properties:
                        id:
                          type: string
                          format: uuid
                          example: a24a6ea4-ce75-4665-a070-57453082c25
                        name:
                          type: string
                          example: EP01
                        project_id:
                          type: string
                          format: uuid
                          example: b24a6ea4-ce75-4665-a070-57453082c25
                        description:
                          type: string
                          example: A short description of the episode
                        status:
                          type: string
                          example: running
        """
        episode = shots_service.get_full_episode(episode_id)
        user_service.check_project_access(episode["project_id"])
        return episode

    @jwt_required()
    def delete(self, episode_id):
        """
        Delete episode
        ---
        tags:
        - Shots
        description: Delete an episode by id. Requires manager access or
          ownership.
        parameters:
          - in: path
            name: episode_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            204:
                description: Episode deleted
        """
        force = self.get_force()
        episode = shots_service.get_episode(episode_id)
        if episode["created_by"] == persons_service.get_current_user()["id"]:
            user_service.check_belong_to_project(episode["project_id"])
        else:
            user_service.check_manager_project_access(episode["project_id"])
        deletion_service.remove_episode(episode_id, force=force)
        return "", 204


class EpisodesResource(Resource):
    @jwt_required()
    def get(self):
        """
        Get episodes
        ---
        tags:
        - Shots
        description: Get episodes with optional filters. Use project_id to
          filter by project.
        parameters:
          - in: query
            name: project_id
            required: False
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: All episode entries
                content:
                  application/json:
                    schema:
                      type: array
                      items:
                        type: object
                        properties:
                          id:
                            type: string
                            format: uuid
                            example: a24a6ea4-ce75-4665-a070-57453082c25
                          name:
                            type: string
                            example: EP01
                          project_id:
                            type: string
                            format: uuid
                            example: b24a6ea4-ce75-4665-a070-57453082c25
                          description:
                            type: string
                            example: A short description of the episode
                          status:
                            type: string
                            example: running
        """
        criterions = query.get_query_criterions_from_request(request)
        user_service.check_project_access(criterions.get("project_id", None))
        return shots_service.get_episodes(criterions)


class EpisodeSequencesResource(Resource):
    @jwt_required()
    def get(self, episode_id):
        """
        Get episode sequences
        ---
        tags:
        - Shots
        description: Get sequences for an episode. You can add query filters
          if needed.
        parameters:
          - in: path
            name: episode_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: query
            name: project_id
            required: False
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: All sequence entries for given episode
                content:
                  application/json:
                    schema:
                      type: array
                      items:
                        type: object
                        properties:
                          id:
                            type: string
                            format: uuid
                            example: a24a6ea4-ce75-4665-a070-57453082c25
                          name:
                            type: string
                            example: SQ01
                          project_id:
                            type: string
                            format: uuid
                            example: b24a6ea4-ce75-4665-a070-57453082c25
                          parent_id:
                            type: string
                            format: uuid
                            example: c24a6ea4-ce75-4665-a070-57453082c25
        """
        if not fields.is_valid_id(episode_id):
            return []
        episode = shots_service.get_episode(episode_id)
        user_service.check_project_access(episode["project_id"])
        criterions = query.get_query_criterions_from_request(request)
        criterions["parent_id"] = episode_id
        if permissions.has_vendor_permissions():
            return shots_service.get_sequences_for_episode(
                episode_id, only_assigned=True
            )
        else:
            return shots_service.get_sequences(criterions)


class EpisodeTaskTypesResource(Resource):
    @jwt_required()
    def get(self, episode_id):
        """
        Get episode task types
        ---
        tags:
        - Shots
        parameters:
          - in: path
            name: episode_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: All task types related to given episode
                content:
                  application/json:
                    schema:
                      type: array
                      items:
                        type: object
                        properties:
                          id:
                            type: string
                            format: uuid
                            example: a24a6ea4-ce75-4665-a070-57453082c25
                          name:
                            type: string
                            example: Animation
        """
        episode = shots_service.get_episode(episode_id)
        user_service.check_project_access(episode["project_id"])
        return tasks_service.get_task_types_for_episode(episode_id)


class EpisodeTasksResource(Resource):
    @jwt_required()
    def get(self, episode_id):
        """
        Get episode tasks
        ---
        tags:
        - Shots
        parameters:
          - in: path
            name: episode_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: All tasks related to given episode
                content:
                  application/json:
                    schema:
                      type: array
                      items:
                        type: object
                        properties:
                          id:
                            type: string
                            format: uuid
                            example: a24a6ea4-ce75-4665-a070-57453082c25
                          name:
                            type: string
                            example: EP01 Layout
                          task_type_id:
                            type: string
                            format: uuid
                            example: b24a6ea4-ce75-4665-a070-57453082c25
                          task_status_id:
                            type: string
                            format: uuid
                            example: c24a6ea4-ce75-4665-a070-57453082c25
                          entity_id:
                            type: string
                            format: uuid
                            example: d24a6ea4-ce75-4665-a070-57453082c25
                          project_id:
                            type: string
                            format: uuid
                            example: e24a6ea4-ce75-4665-a070-57453082c25
                          assignees:
                            type: array
                            items:
                              type: string
                              format: uuid
                            example: ["f24a6ea4-ce75-4665-a070-57453082c25"]
        """
        episode = shots_service.get_episode(episode_id)
        user_service.check_project_access(episode["project_id"])
        return tasks_service.get_tasks_for_episode(episode_id)


class SequenceResource(Resource, ArgsMixin):
    @jwt_required()
    def get(self, sequence_id):
        """
        Get sequence
        ---
        tags:
        - Shots
        description: Get a sequence by id.
        parameters:
          - in: path
            name: sequence_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: Sequence found and returned
                content:
                  application/json:
                    schema:
                      type: object
                      properties:
                        id:
                          type: string
                          format: uuid
                          example: a24a6ea4-ce75-4665-a070-57453082c25
                        name:
                          type: string
                          example: SQ01
                        project_id:
                          type: string
                          format: uuid
                          example: b24a6ea4-ce75-4665-a070-57453082c25
                        parent_id:
                          type: string
                          format: uuid
                          example: c24a6ea4-ce75-4665-a070-57453082c25
        """
        sequence = shots_service.get_full_sequence(sequence_id)
        user_service.check_project_access(sequence["project_id"])
        return sequence

    @jwt_required()
    def delete(self, sequence_id):
        """
        Delete sequence
        ---
        tags:
        - Shots
        description: Delete a sequence by id. Requires manager access or
          ownership.
        parameters:
          - in: path
            name: sequence_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            204:
                description: Sequence deleted
        """
        force = self.get_force()
        sequence = shots_service.get_sequence(sequence_id)
        if sequence["created_by"] != persons_service.get_current_user()["id"]:
            user_service.check_belong_to_project(sequence["project_id"])
        else:
            user_service.check_manager_project_access(sequence["project_id"])
        shots_service.remove_sequence(sequence_id, force=force)
        return "", 204


class SequencesResource(Resource):
    @jwt_required()
    def get(self):
        """
        Get sequences
        ---
        tags:
        - Shots
        description: Get sequences with optional filters. Use episode_id to
          filter by episode.
        parameters:
          - in: query
            name: episode_id
            required: False
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: All sequence entries
                content:
                  application/json:
                    schema:
                      type: array
                      items:
                        type: object
                        properties:
                          id:
                            type: string
                            format: uuid
                            example: a24a6ea4-ce75-4665-a070-57453082c25
                          name:
                            type: string
                            example: SQ01
                          project_id:
                            type: string
                            format: uuid
                            example: b24a6ea4-ce75-4665-a070-57453082c25
                          parent_id:
                            type: string
                            format: uuid
                            example: c24a6ea4-ce75-4665-a070-57453082c25
        """
        criterions = query.get_query_criterions_from_request(request)
        if "episode_id" in criterions:
            episode = shots_service.get_episode(criterions["episode_id"])
            criterions["project_id"] = episode["project_id"]
            criterions["parent_id"] = episode["id"]
            del criterions["episode_id"]
        user_service.check_project_access(criterions.get("project_id", None))
        return shots_service.get_sequences(criterions)


class SequenceShotsResource(Resource):
    @jwt_required()
    def get(self, sequence_id):
        """
        Get sequence shots
        ---
        tags:
        - Shots
        description: Get shots for a sequence. Supports filtering using query
          params.
        parameters:
          - in: path
            name: sequence_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: query
            name: project_id
            required: False
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: All shot entries for given sequence
                content:
                  application/json:
                    schema:
                      type: array
                      items:
                        type: object
                        properties:
                          id:
                            type: string
                            format: uuid
                            example: a24a6ea4-ce75-4665-a070-57453082c25
                          name:
                            type: string
                            example: SH010
                          project_id:
                            type: string
                            format: uuid
                            example: b24a6ea4-ce75-4665-a070-57453082c25
                          parent_id:
                            type: string
                            format: uuid
                            example: c24a6ea4-ce75-4665-a070-57453082c25
        """
        sequence = shots_service.get_sequence(sequence_id)
        user_service.check_project_access(sequence["project_id"])
        criterions = query.get_query_criterions_from_request(request)
        criterions["parent_id"] = sequence_id
        if permissions.has_vendor_permissions():
            criterions["assigned_to"] = persons_service.get_current_user()[
                "id"
            ]
        return shots_service.get_shots(criterions)


class ProjectScenesResource(Resource, ArgsMixin):
    @jwt_required()
    def get(self, project_id):
        """
        Get project scenes
        ---
        tags:
        - Shots
        description: Get all scenes for a project.
        parameters:
          - in: path
            name: project_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: All scenes related to given project
                content:
                  application/json:
                    schema:
                      type: array
                      items:
                        type: object
                        properties:
                          id:
                            type: string
                            format: uuid
                            example: a24a6ea4-ce75-4665-a070-57453082c25
                          name:
                            type: string
                            example: SC001
                          project_id:
                            type: string
                            format: uuid
                            example: b24a6ea4-ce75-4665-a070-57453082c25
                          parent_id:
                            type: string
                            format: uuid
                            example: c24a6ea4-ce75-4665-a070-57453082c25
        """
        projects_service.get_project(project_id)
        user_service.check_project_access(project_id)
        return shots_service.get_scenes_for_project(project_id)

    @jwt_required()
    def post(self, project_id):
        """
        Create project scene
        ---
        tags:
        - Shots
        description: Create a new scene in a project. Provide a name and the
          related sequence id.
        parameters:
          - in: path
            name: project_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: object
                required:
                  - name
                  - sequence_id
                properties:
                  name:
                    type: string
                    required: true
                    example: "Name of scene"
                  sequence_id:
                    type: string
                    format: uuid
                    required: true
                    example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            201:
                description: Scene created for given project
                content:
                  application/json:
                    schema:
                      type: object
                      properties:
                        id:
                          type: string
                          format: uuid
                          example: a24a6ea4-ce75-4665-a070-57453082c25
                        name:
                          type: string
                          example: SC001
                        project_id:
                          type: string
                          format: uuid
                          example: b24a6ea4-ce75-4665-a070-57453082c25
                        parent_id:
                          type: string
                          format: uuid
                          example: c24a6ea4-ce75-4665-a070-57453082c25
        """
        (sequence_id, name) = self.get_arguments()
        projects_service.get_project(project_id)
        user_service.check_manager_project_access(project_id)
        scene = shots_service.create_scene(
            project_id,
            sequence_id,
            name,
            created_by=persons_service.get_current_user()["id"],
        )
        return scene, 201

    def get_arguments(self):
        args = self.get_args(
            [{"name": "name", "required": True}, "sequence_id"]
        )

        return (args["sequence_id"], args["name"])


class SequenceScenesResource(Resource):
    @jwt_required()
    def get(self, sequence_id):
        """
        Get sequence scenes
        ---
        tags:
        - Shots
        description: Get scenes that belong to a sequence.
        parameters:
          - in: path
            name: sequence_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: All scenes related to given sequence
                content:
                  application/json:
                    schema:
                      type: array
                      items:
                        type: object
                        properties:
                          id:
                            type: string
                            format: uuid
                            example: a24a6ea4-ce75-4665-a070-57453082c25
                          name:
                            type: string
                            example: SC010
                          project_id:
                            type: string
                            format: uuid
                            example: b24a6ea4-ce75-4665-a070-57453082c25
                          parent_id:
                            type: string
                            format: uuid
                            example: c24a6ea4-ce75-4665-a070-57453082c25
        """
        sequence = shots_service.get_sequence(sequence_id)
        user_service.check_project_access(sequence["project_id"])
        return shots_service.get_scenes_for_sequence(sequence_id)


class SceneTaskTypesResource(Resource):
    @jwt_required()
    def get(self, scene_id):
        """
        Get scene task types
        ---
        tags:
        - Shots
        description: Get task types for a scene.
        parameters:
          - in: path
            name: scene_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: All task types related to given scene
                content:
                  application/json:
                    schema:
                      type: array
                      items:
                        type: object
                        properties:
                          id:
                            type: string
                            format: uuid
                            example: a24a6ea4-ce75-4665-a070-57453082c25
                          name:
                            type: string
                            example: Animation
        """
        scene = shots_service.get_scene(scene_id)
        user_service.check_project_access(scene["project_id"])
        user_service.check_entity_access(scene["id"])
        return tasks_service.get_task_types_for_scene(scene_id)


class SceneTasksResource(Resource):
    @jwt_required()
    def get(self, scene_id):
        """
        Get scene tasks
        ---
        tags:
        - Shots
        description: Get tasks for a scene.
        parameters:
          - in: path
            name: scene_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: All tasks related to given scene
                content:
                  application/json:
                    schema:
                      type: array
                      items:
                        type: object
                        properties:
                          id:
                            type: string
                            format: uuid
                            example: a24a6ea4-ce75-4665-a070-57453082c25
                          name:
                            type: string
                            example: SC001 Layout
                          task_type_id:
                            type: string
                            format: uuid
                            example: b24a6ea4-ce75-4665-a070-57453082c25
                          task_status_id:
                            type: string
                            format: uuid
                            example: c24a6ea4-ce75-4665-a070-57453082c25
                          entity_id:
                            type: string
                            format: uuid
                            example: d24a6ea4-ce75-4665-a070-57453082c25
                          project_id:
                            type: string
                            format: uuid
                            example: e24a6ea4-ce75-4665-a070-57453082c25
        """
        scene = shots_service.get_scene(scene_id)
        user_service.check_entity_access(scene["id"])
        return tasks_service.get_tasks_for_scene(scene_id)


class SceneShotsResource(Resource, ArgsMixin):
    @jwt_required()
    def get(self, scene_id):
        """
        Get scene shots
        ---
        tags:
        - Shots
        description: Get shots that come from a scene.
        parameters:
          - in: path
            name: scene_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: All shots that come from given scene
                content:
                  application/json:
                    schema:
                      type: array
                      items:
                        type: object
                        properties:
                          id:
                            type: string
                            format: uuid
                            example: a24a6ea4-ce75-4665-a070-57453082c25
                          name:
                            type: string
                            example: SH010
                          project_id:
                            type: string
                            format: uuid
                            example: b24a6ea4-ce75-4665-a070-57453082c25
        """
        scene = shots_service.get_scene(scene_id)
        user_service.check_entity_access(scene["id"])
        return scenes_service.get_shots_by_scene(scene_id)

    @jwt_required()
    def post(self, scene_id):
        """
        Link shot to scene
        ---
        tags:
        - Shots
        description: Link a shot to a scene as its source.
        parameters:
          - in: path
            name: scene_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: object
                required:
                  - shot_id
                properties:
                  shot_id:
                    type: string
                    format: uuid
                    required: true
                    example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            201:
                description: Scene marked as source of shot
                content:
                  application/json:
                    schema:
                      type: object
                      properties:
                        scene_id:
                          type: string
                          format: uuid
                          example: a24a6ea4-ce75-4665-a070-57453082c25
                        shot_id:
                          type: string
                          format: uuid
                          example: b24a6ea4-ce75-4665-a070-57453082c25
        """
        args = self.get_args([("shot_id", None, True)])

        scene = shots_service.get_scene(scene_id)
        user_service.check_project_access(scene["project_id"])
        shot = shots_service.get_shot(args["shot_id"])
        return scenes_service.add_shot_to_scene(scene, shot), 201


class RemoveShotFromSceneResource(Resource):
    @jwt_required()
    def delete(self, scene_id, shot_id):
        """
        Delete given shot from given scene.
        ---
        tags:
        - Shots
        parameters:
          - in: path
            name: scene_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: shot_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            204:
                description: Given shot deleted from given scene
        """
        scene = shots_service.get_scene(scene_id)
        user_service.check_project_access(scene["project_id"])
        shot = shots_service.get_shot(shot_id)
        scenes_service.remove_shot_from_scene(scene, shot)
        return "", 204


class ShotVersionsResource(Resource):
    """
    Get shot versions
    """

    @jwt_required()
    def get(self, shot_id):
        """
        Get shot versions
        ---
        tags:
        - Shots
        description: Get data versions of a shot. Use this to inspect version
          history.
        parameters:
          - in: path
            name: shot_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: Data versions of given shot
                content:
                  application/json:
                    schema:
                      type: array
                      items:
                        type: object
                        properties:
                          id:
                            type: string
                            format: uuid
                            example: a24a6ea4-ce75-4665-a070-57453082c25
                          created_at:
                            type: string
                            format: date-time
                            example: "2024-01-15T10:30:00Z"
        """
        shot = shots_service.get_shot(shot_id)
        user_service.check_project_access(shot["project_id"])
        user_service.check_entity_access(shot["id"])
        return shots_service.get_shot_versions(shot_id)


class ProjectQuotasResource(Resource, ArgsMixin):

    @jwt_required()
    def get(self, project_id, task_type_id):
        """
        Get project quotas
        ---
        tags:
        - Shots
        description: Get quotas statistics for a project and task type. Supports
          weighted and raw modes with optional feedback filtering.
        parameters:
          - in: path
            name: project_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: task_type_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: query
            name: count_mode
            required: True
            type: string
            enum: [weighted, weigtheddone, feedback, done]
            example: weighted
          - in: query
            name: studio_id
            required: False
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: Quotas statistics for shots
                content:
                  application/json:
                    schema:
                      type: object
                      additionalProperties:
                        type: object
                        properties:
                          count:
                            type: integer
                            example: 42
                          frames:
                            type: integer
                            example: 1200
            400:
                description: Invalid count_mode or parameter
        """
        projects_service.get_project(project_id)
        user_service.check_project_access(project_id)
        args = self.get_args(
            [
                ("count_mode", "weighted", False, str),
                ("studio_id", None, False, str),
            ]
        )
        count_mode = args["count_mode"]
        studio_id = args["studio_id"]

        if count_mode not in ["weighted", "weighteddone", "feedback", "done"]:
            raise WrongParameterException(
                "count_mode must be equal to weighted, weigtheddone, feedback"
                ", or done"
            )

        feedback = "done" not in count_mode
        weighted = "weighted" in count_mode

        if weighted:
            return shots_service.get_weighted_quotas(
                project_id,
                task_type_id,
                feedback=feedback,
                studio_id=studio_id,
            )
        else:
            return shots_service.get_raw_quotas(
                project_id,
                task_type_id,
                feedback=feedback,
                studio_id=studio_id,
            )


class ProjectPersonQuotasResource(Resource, ArgsMixin):

    @jwt_required()
    def get(self, project_id, person_id):
        """
        Get project person quotas
        ---
        tags:
        - Shots
        description: Get quotas statistics for a person in a project. Supports
          weighted and raw modes with optional feedback filtering.
        parameters:
          - in: path
            name: project_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: person_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: query
            name: count_mode
            required: True
            type: string
            enum: [weighted, weigtheddone, feedback, done]
            example: weighted
          - in: query
            name: studio_id
            required: False
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: Quotas statistics for shots
                content:
                  application/json:
                    schema:
                      type: object
                      additionalProperties:
                        type: object
                        properties:
                          count:
                            type: integer
                            example: 15
                          frames:
                            type: integer
                            example: 360
            400:
                description: Invalid count_mode or parameter
        """
        projects_service.get_project(project_id)
        if (
            permissions.has_manager_permissions()
            or permissions.has_supervisor_permissions()
        ):
            user_service.check_project_access(project_id)
        else:
            user_service.check_person_access(person_id)
        args = self.get_args(
            [
                ("count_mode", "weighted", False, str),
                ("studio_id", None, False, str),
            ]
        )
        count_mode = args["count_mode"]
        studio_id = args["studio_id"]

        if count_mode not in ["weighted", "weighteddone", "feedback", "done"]:
            raise WrongParameterException(
                "count_mode must be equal to weighted, weigtheddone, feedback"
                ", or done"
            )

        feedback = "done" not in count_mode
        weighted = "weighted" in count_mode

        if weighted:
            return shots_service.get_weighted_quotas(
                project_id,
                person_id=person_id,
                feedback=feedback,
                studio_id=studio_id,
            )
        else:
            return shots_service.get_raw_quotas(
                project_id,
                person_id=person_id,
                feedback=feedback,
                studio_id=studio_id,
            )


class SetShotsFramesResource(Resource, ArgsMixin):
    @jwt_required()
    def post(self, project_id, task_type_id):
        """
        Set shots frames
        ---
        tags:
        - Shots
        description: Set number of frames on shots based on latest preview
          files for a task type. Optionally scope by episode via query
          param.
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: object
                required:
                  - shots
                properties:
                  shots:
                    type: array
                    required: true
                    items:
                      type: object
                      properties:
                        shot_id:
                          type: string
                          format: uuid
                          example: a24a6ea4-ce75-4665-a070-57453082c25
                        nb_frames:
                          type: integer
                          example: 24
        responses:
            200:
                description: Frames set for given shots
                content:
                  application/json:
                    schema:
                      type: object
                      properties:
                        updated:
                          type: integer
                          example: 12
            400:
                description: Invalid ids or parameters
        """
        user_service.check_manager_project_access(project_id)
        if not fields.is_valid_id(task_type_id) or not fields.is_valid_id(
            project_id
        ):
            raise WrongParameterException("Invalid project or task type id")

        episode_id = self.get_episode_id()
        if not episode_id in ["", None] and not fields.is_valid_id(episode_id):
            raise WrongParameterException("Invalid episode id")

        if episode_id == "":
            episode_id = None

        return shots_service.set_frames_from_task_type_preview_files(
            project_id,
            task_type_id,
            episode_id=episode_id,
        )



from flask import current_app
from flask_jwt_extended import jwt_required

from zou.app.models.project import Project
from zou.app.models.project_status import ProjectStatus

from zou.app.services import file_tree_service, projects_service

from zou.app.blueprints.source.shotgun.base import (
    BaseImportShotgunResource,
    ImportRemoveShotgunBaseResource,
)

from zou.app.services.exception import WrongFileTreeFileException


class ImportShotgunProjectsResource(BaseImportShotgunResource):
    def __init__(self):
        BaseImportShotgunResource.__init__(self)

    @jwt_required()
    def post(self):
        """
        Import shotgun projects
        ---
        description: Import Shotgun projects. Send a list of Shotgun project
          entries in the JSON body. Returns created or updated projects with
          custom fields preserved.
        tags:
          - Import
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: array
                items:
                  type: object
                  properties:
                    id:
                      type: integer
                      description: Shotgun ID of the project
                      example: 12345
                    name:
                      type: string
                      description: Project name
                      example: "My Project"
                    sg_status:
                      type: string
                      description: Project status
                      example: "Active"
              example:
                - id: 12345
                  name: "My Project"
                  sg_status: "Active"
        responses:
          200:
            description: Projects imported successfully
            content:
              application/json:
                schema:
                  type: array
                  items:
                    type: object
                    properties:
                      id:
                        type: string
                        format: uuid
                        description: Project unique identifier
                        example: a24a6ea4-ce75-4665-a070-57453082c25
                      name:
                        type: string
                        description: Project name
                        example: "My Project"
                      data:
                        type: object
                        description: Custom project data
                        example: {}
                      created_at:
                        type: string
                        format: date-time
                        description: Creation timestamp
                        example: "2024-01-15T10:30:00Z"
                      updated_at:
                        type: string
                        format: date-time
                        description: Update timestamp
                        example: "2024-01-15T11:00:00Z"
          400:
            description: Invalid request body or data format error
        """
        return super().post()

    def prepare_import(self):
        self.project_status_names = self.extract_status_names(self.sg_entries)
        projects_service.save_project_status(self.project_status_names)
        self.project_status_map = ProjectStatus.get_id_map(field="name")

    def extract_status_names(self, sg_projects):
        return {x["sg_status"] for x in sg_projects}

    def extract_data(self, sg_project):
        sg_project_status = sg_project["sg_status"]
        custom_fields = self.extract_custom_data(sg_project)
        project_status_id = self.project_status_map.get(
            sg_project_status, None
        )

        data = {
            "project_status_id": project_status_id,
            "name": sg_project["name"],
            "shotgun_id": sg_project["id"],
            "data": {},
        }

        data["data"] = custom_fields
        return data

    def is_custom_field(self, name):
        non_custom_fields = ["sg_status"]
        return name[:3] == "sg_" and name not in non_custom_fields

    def import_entry(self, data):
        project = Project.get_by(shotgun_id=data["shotgun_id"])

        if project is None:
            project = Project(**data)

            tree_name = current_app.config["DEFAULT_FILE_TREE"]
            try:
                project.file_tree = file_tree_service.get_tree_from_file(
                    tree_name
                )
            except WrongFileTreeFileException:
                current_app.logger.error(
                    "Can't find default file to set project file tree. Set "
                    "an empty file tree instead."
                )
                project.file_tree = {}

            project.save()
            current_app.logger.info("Project created: %s" % project)

        else:
            project.update(data)
            current_app.logger.info("Project updated: %s" % project)

        return project


class ImportRemoveShotgunProjectResource(ImportRemoveShotgunBaseResource):
    def __init__(self):
        ImportRemoveShotgunBaseResource.__init__(self, Project)

    @jwt_required()
    def post(self):
        """
        Remove shotgun project
        ---
        description: Remove a Shotgun project from the database. Provide the
          Shotgun entry ID in the JSON body.
        tags:
          - Import
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: object
                required:
                  - id
                properties:
                  id:
                    type: integer
                    description: Shotgun ID of the project to remove
                    example: 12345
              example:
                id: 12345
        responses:
          200:
            description: Removal result returned
            content:
              application/json:
                schema:
                  type: object
                  properties:
                    success:
                      type: boolean
                      description: Whether the removal was successful
                      example: true
                    removed_instance_id:
                      type: string
                      format: uuid
                      description: ID of the removed project, if found
                      example: a24a6ea4-ce75-4665-a070-57453082c25
          400:
            description: Invalid request body or instance not found
        """
        return super().post()



import email.utils
import spdx_license_list
import zipfile
import importlib
import importlib.util
import sys
import os
import tomlkit
import traceback
import semver
import shutil
import subprocess
import tempfile

from alembic import command
from alembic.config import Config
from collections.abc import MutableMapping
from flask import Blueprint, current_app
from flask_restful import Resource
from pathlib import Path

from zou.app.utils.api import configure_api_from_blueprint

from flask import send_from_directory, abort, current_app


class StaticResource(Resource):

    plugin_id = None

    def get(self, filename="index.html"):

        print(self.plugin_id)
        static_folder = (
            Path(current_app.config.get("PLUGIN_FOLDER", "plugins"))
            / self.plugin_id
            / "frontend"
            / "dist"
        )

        if filename == "":
            filename = "index.html"

        file_path = static_folder / filename
        if not file_path.exists() or not file_path.is_file():
            abort(404)

        if filename == "":
            filename = "index.html"

        return send_from_directory(
            str(static_folder), filename, conditional=True, max_age=0
        )


class PluginManifest(MutableMapping):
    def __init__(self, data):
        super().__setattr__("data", data)
        self.validate()

    @classmethod
    def from_plugin_path(cls, path):
        path = Path(path)
        if path.is_dir():
            return cls.from_file(path / "manifest.toml")
        elif zipfile.is_zipfile(path):
            with zipfile.ZipFile(path) as z:
                with z.open("manifest.toml") as f:
                    data = tomlkit.load(f)
            return cls(data)
        else:
            raise ValueError(f"Invalid plugin path: {path}")

    @classmethod
    def from_file(cls, path):
        with open(path, "rb") as f:
            data = tomlkit.load(f)
        return cls(data)

    def write_to_path(self, path):
        path = Path(path)
        with open(path / "manifest.toml", "w", encoding="utf-8") as f:
            tomlkit.dump(self.data, f)

    def validate(self):
        semver.Version.parse(str(self.data["version"]))
        spdx_license_list.LICENSES[self.data["license"]]
        if "maintainer" in self.data:
            name, email_addr = email.utils.parseaddr(self.data["maintainer"])
            self.data["maintainer_name"] = name
            self.data["maintainer_email"] = email_addr

        if "frontend_project_enabled" not in self.data:
            self.data["frontend_project_enabled"] = False
        if "frontend_studio_enabled" not in self.data:
            self.data["frontend_studio_enabled"] = False

    def to_model_dict(self):
        return {
            "plugin_id": self.data["id"],
            "name": self.data["name"],
            "description": self.data.get("description"),
            "version": str(self.data["version"]),
            "maintainer_name": self.data.get("maintainer_name"),
            "maintainer_email": self.data.get("maintainer_email"),
            "website": self.data.get("website"),
            "license": self.data["license"],
            "frontend_project_enabled": self.data.get(
                "frontend_project_enabled", False
            ),
            "frontend_studio_enabled": self.data.get(
                "frontend_studio_enabled", False
            ),
            "icon": self.data.get("icon", ""),
        }

    def __getitem__(self, key):
        return self.data[key]

    def __setitem__(self, key, value):
        self.data[key] = value

    def __delitem__(self, key):
        del self.data[key]

    def __iter__(self):
        return iter(self.data)

    def __len__(self):
        return len(self.data)

    def __repr__(self):
        return f"<PluginManifest {self.data!r}>"

    def __getattr__(self, attr):
        try:
            return self.data[attr]
        except KeyError:
            raise AttributeError(f"'PluginManifest' has no attribute '{attr}'")

    def __setattr__(self, attr, value):
        if attr == "data":
            super().__setattr__(attr, value)
        else:
            self.data[attr] = value


def load_plugin(app, plugin_path, init_plugin=True):
    """
    Load a plugin from the plugin folder.
    """
    plugin_path = Path(plugin_path)
    manifest = PluginManifest.from_plugin_path(plugin_path)
    plugin_module = importlib.import_module(manifest["id"])

    if not hasattr(plugin_module, "routes"):
        raise Exception(f"Plugin {manifest['id']} has no routes.")

    routes = plugin_module.routes
    add_static_routes(manifest, routes)
    blueprint = Blueprint(manifest["id"], manifest["id"])
    configure_api_from_blueprint(blueprint, routes)
    app.register_blueprint(blueprint, url_prefix=f"/plugins/{manifest['id']}")

    return plugin_module


def load_plugins(app):
    """
    Load plugins from the plugin folder.
    """
    plugin_folder = Path(app.config["PLUGIN_FOLDER"])
    if plugin_folder.exists():
        abs_plugin_path = str(plugin_folder.absolute())
        if abs_plugin_path not in sys.path:
            sys.path.insert(0, abs_plugin_path)

        for plugin_id in os.listdir(plugin_folder):
            try:
                load_plugin(app, plugin_folder / plugin_id)
                app.logger.info(f"Plugin {plugin_id} loaded.")
            except ImportError as e:
                app.logger.error(f"Plugin {plugin_id} failed to import: {e}")
            except Exception as e:
                app.logger.error(
                    f"Plugin {plugin_id} failed to initialize: {e}"
                )
                app.logger.debug(traceback.format_exc())

        if abs_plugin_path in sys.path:
            sys.path.remove(abs_plugin_path)


def migrate_plugin_db(plugin_path, message):
    """
    Generates Alembic migration files in path/migrations.
    """
    plugin_path = Path(plugin_path).absolute()
    models_path = plugin_path / "models.py"

    if not models_path.exists():
        raise FileNotFoundError(f"'models.py' not found in '{plugin_path}'")

    manifest = PluginManifest.from_plugin_path(plugin_path)

    module_name = f"_plugin_models_{manifest['id']}"
    spec = importlib.util.spec_from_file_location(module_name, models_path)
    if spec is None or spec.loader is None:
        raise ImportError(f"Could not load 'models.py' from '{plugin_path}'")

    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    try:
        spec.loader.exec_module(module)
        migrations_dir = plugin_path / "migrations"
        versions_dir = migrations_dir / "versions"
        versions_dir.mkdir(parents=True, exist_ok=True)

        alembic_cfg = Config()
        alembic_cfg.config_file_name = str(
            plugin_path / "migrations" / "alembic.ini"
        )
        alembic_cfg.set_main_option("script_location", str(migrations_dir))
        alembic_cfg.set_main_option(
            "sqlalchemy.url", current_app.config["SQLALCHEMY_DATABASE_URI"]
        )

        command.revision(alembic_cfg, autogenerate=True, message=message)
    finally:
        del sys.modules[module_name]


def run_plugin_migrations(plugin_path, plugin):
    """
    Run plugin migrations.
    """
    plugin_path = Path(plugin_path)

    alembic_cfg = Config()
    alembic_cfg.config_file_name = str(
        plugin_path / "migrations" / "alembic.ini"
    )
    alembic_cfg.set_main_option(
        "script_location", str(plugin_path / "migrations")
    )
    alembic_cfg.set_main_option(
        "sqlalchemy.url", current_app.config["SQLALCHEMY_DATABASE_URI"]
    )

    command.upgrade(alembic_cfg, "head")

    script = command.ScriptDirectory.from_config(alembic_cfg)
    head_revision = script.get_current_head()

    plugin.revision = head_revision

    return head_revision


def downgrade_plugin_migrations(plugin_path):
    """
    Downgrade plugin migrations to base.
    """
    plugin_path = Path(plugin_path)
    manifest = PluginManifest.from_plugin_path(plugin_path)

    alembic_cfg = Config()
    alembic_cfg.config_file_name = str(
        plugin_path / "migrations" / "alembic.ini"
    )
    alembic_cfg.set_main_option(
        "script_location", str(plugin_path / "migrations")
    )
    alembic_cfg.set_main_option(
        "sqlalchemy.url", current_app.config["SQLALCHEMY_DATABASE_URI"]
    )

    try:
        command.downgrade(alembic_cfg, "base")
    except Exception as e:
        current_app.logger.warning(
            f"Downgrade failed for plugin {manifest.id}: {e}"
        )


def create_plugin_package(path, output_path, force=False):
    """
    Create a plugin package.
    """
    path = Path(path)
    if not path.exists():
        raise FileNotFoundError(f"Plugin path '{path}' does not exist.")

    manifest = PluginManifest.from_plugin_path(path)

    output_path = Path(output_path)
    if not output_path.suffix == ".zip":
        output_path /= f"{manifest.id}-{manifest.version}.zip"
    if output_path.exists():
        if force:
            output_path.unlink()
        else:
            raise FileExistsError(
                f"Output path '{output_path}' already exists."
            )

    output_path = shutil.make_archive(
        output_path.with_suffix(""),
        "zip",
        path,
    )
    return output_path


def create_plugin_skeleton(
    path,
    id,
    name,
    description=None,
    version=None,
    maintainer=None,
    website=None,
    license=None,
    icon=None,
    force=False,
):
    plugin_template_path = (
        Path(__file__).parent.parent.parent / "plugin_template"
    )
    plugin_path = Path(path) / id

    if plugin_path.exists():
        if force:
            shutil.rmtree(plugin_path)
        else:
            raise FileExistsError(
                f"Plugin '{id}' already exists in {plugin_path}."
            )

    shutil.copytree(plugin_template_path, plugin_path)

    manifest = PluginManifest.from_file(plugin_path / "manifest.toml")

    manifest.id = id
    manifest.name = name
    if description:
        manifest.description = description
    if version:
        manifest.version = version
    if maintainer:
        manifest.maintainer = maintainer
    if website:
        manifest.website = website
    if license:
        manifest.license = license
    if icon:
        manifest.icon = icon
    manifest.validate()
    manifest.write_to_path(plugin_path)

    return plugin_path


def install_plugin_files(files_path, installation_path):
    """
    Install plugin files.
    """
    files_path = Path(files_path)
    installation_path = Path(installation_path)

    installation_path.mkdir(parents=True, exist_ok=True)

    if files_path.is_dir():
        def ignore_git(dir, names):
            ignored = []
            if ".git" in names:
                ignored.append(".git")
            return ignored
        shutil.copytree(
            files_path, installation_path, dirs_exist_ok=True, ignore=ignore_git
        )
    elif zipfile.is_zipfile(files_path):
        shutil.unpack_archive(files_path, installation_path, format="zip")
    else:
        raise ValueError(
            f"Plugin path '{files_path}' is not a valid zip file or a directory."
        )

    return installation_path


def uninstall_plugin_files(plugin_path):
    """
    Uninstall plugin files.
    """
    plugin_path = Path(plugin_path)
    if plugin_path.exists():
        shutil.rmtree(plugin_path)
        return True
    return False


def clone_git_repo(git_url, temp_dir=None):
    """
    Clone a git repository to a temporary directory.
    Returns the path to the cloned directory.
    """
    if temp_dir is None:
        temp_dir = tempfile.mkdtemp(prefix="zou_plugin_")

    temp_dir = Path(temp_dir)
    repo_name = git_url.rstrip("/").split("/")[-1].replace(".git", "")
    clone_path = temp_dir / repo_name

    print(f"[Plugins] Cloning {git_url}...")

    try:
        subprocess.run(
            ["git", "clone", git_url, str(clone_path)],
            check=True,
            capture_output=True,
            timeout=300,
        )
        print(f"[Plugins] Successfully cloned {git_url}")
        return clone_path
    except subprocess.CalledProcessError as e:
        error_msg = e.stderr.decode() if e.stderr else str(e)
        raise ValueError(f"Failed to clone repository {git_url}: {error_msg}")
    except FileNotFoundError:
        raise ValueError(
            "git is not available. Please install git to clone repositories."
        )


def add_static_routes(manifest, routes):
    """
    Add static routes to the manifest.
    """

    class PluginStaticResource(StaticResource):

        def __init__(self):
            self.plugin_id = manifest.id
            super().__init__()

    if (
        manifest["frontend_project_enabled"]
        or manifest["frontend_studio_enabled"]
    ):
        routes.append((f"/frontend/<path:filename>", PluginStaticResource))
        routes.append((f"/frontend", PluginStaticResource))
